#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division
import numpy as np
# import sys
# sys.path.insert(0, '/home/did/RTC/SMART-G/')
from smartg.rrs import L2d_inv, is_odd
from smartg.vrs import V2d_inv
from luts.luts import LUT
import scipy.constants as cst
from scipy.interpolate import interp1d

class BandSet(object):
    def __init__(self, wav):
        '''
        Common objet for formatting input bands definition

        Valid inputs:
            * float
            * 1-d array
            * KDIS or REPTRAN IBANDS LIST

        Methods:
            __getitem__: returns wavelength
        '''
        try:
            self.use_reptran_kdis = hasattr(wav[0], 'calc_profile')
        except:
            self.use_reptran_kdis = False

        self.type_wav = False
        if self.use_reptran_kdis:
            self.wav = [x.w for x in wav]
            self.data = wav
            self.type_wav = type(wav[0])
        else:
            self.wav = wav
            self.data = None
            self.type_wav =None
        

        assert isinstance(self.wav, (float, list, np.ndarray))
        self.wav = np.array(self.wav, dtype='float32')
        self.scalar = (self.wav.ndim == 0)
        if self.scalar:
            self.wav = self.wav.reshape(1)
        self.size = self.wav.size

    def __getitem__(self, key):
        return self.wav[key]

    def __len__(self):
        return self.size

    def calc_profile(self, prof):
        '''
        calculate the absorption profile for each band
        '''
        tau_mol = np.zeros((self.size, len(prof.z)), dtype='float32')

        if self.use_reptran_kdis:
            for i, w in enumerate(self.data):
                tau_mol[i,:] = w.calc_profile(prof)

        return tau_mol


def spectral_grids(lmin, lmax, datas, dl=None, dls=None, Raman='RRS', unit='mW/m2/nm'):
    '''
    inputs:
        lmin  : lambda min (nm)
        lmax  : lambda max (nm)
        datas : solar spectrum data
    
    keywords:
        dl : high spectral resolution for absorption features # nm (default, None, same as datas)
        dls: low  spectral resolution for scattering features # nm (default None), NWS=1
        Raman: 'RRS' or 'VRS'
        unit: solar irradiance unit 'photons/cm2/s/nm' or 'mW/m2/nm'
    '''
    ## Solar spectrum input data ##
    wl0 = datas[:,0]
    E0  = datas[:,1]
    # from mW/m2/nm to photons/cm2/s/nm
    if (unit=='photons/cm2/s/nm') : E0  *= 1e-3 * 1e-4  / (cst.h *cst.c)  * (wl0*1e-9)
    
    ## High spectral resolution grid (for absorption features)
    if dl is None:
        # Solar grid
        ii = np.where(( wl0>=lmin) & (wl0 <=lmax))
        wl = wl0[ii]
        NW = wl.size
        # Solar resolution
        dl = (wl[-1]-wl[0])/wl.size
    else:
        NW  = int((lmax-lmin)/dl) + 1
        wl  = np.linspace(lmin, lmax, num=NW) # wavelength grid
        
    if Raman=='RRS':
        ## RRS excitation wavelength grid for scattering angle of 90 deg and 243°K
        wl_RS, _ = L2d_inv(wl, 90., 243.)
    else:
        ## VRS excitation wavelength grid
        wl_RS, _ = V2d_inv(wl)

    lmin_RS  = min(wl_RS.min(), wl.min())
    lmax_RS  = max(wl_RS.max(), wl.max())
    
    # Solar spectrum LUT building
    ii = np.where(( wl0>=lmin_RS) & (wl0 <=lmax_RS))
    Es_LUT = LUT(E0[ii], axes=[wl0[ii]], names=['wavelength'], desc='Es')
    
    # low spectral resolution for scattering computations # nm 
    if dls is not None:
        NWS = int((lmax_RS - lmin_RS)/dls) 
        NWS = NWS if is_odd(NWS) else NWS+1
    else : NWS=1
    wls = np.linspace(lmin_RS, lmax_RS, num=NWS)
    
    # parameters for 1D linear interpolation of wl in wls
    if NWS>1:
        f  =  interp1d(wls,np.linspace(0, NWS-1, num=NWS))
        iw =  f(wl)
        iwls_in = np.floor(iw).astype(np.int8)        # index of lower wls value in the wls array, 
        wwls_in = (iw-iwls_in).astype(np.float32)  # floating proportion between iwls and iwls+1
        # special case for NWS
        ii = np.where(iwls_in==(NWS-1))
        iwls_in[ii] = NWS-2
        wwls_in[ii] = 1.
    else:
        iwls_in = np.array([0], dtype=np.int8)
        wwls_in = np.array([0], dtype=np.float32)
    
    return wl, wls, wl_RS, Es_LUT, iwls_in, wwls_in



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
SMART-G test suite using pytest
'''

import pytest
import numpy as np
from smartg.smartg import Smartg, RoughSurface, LambSurface, Albedo_cst
from smartg.atmosphere import AtmAFGL, AerOPAC, Cloud
from smartg.water import IOP_1
from smartg.reptran import REPTRAN, reduce_reptran
from smartg.tools.smartg_view import smartg_view
import os
from smartg import conftest

NBPHOTONS = 1e4

wav_list = [500.,
            [500.],
            np.array([400., 600.])
            ]

@pytest.fixture(params=[True, False])
def sg(request):
    '''
    A fixture to create multiple Smartg instances (pp or spherical)
    '''
    return Smartg(pp=request.param)


@pytest.mark.parametrize('pp', [True, False])
@pytest.mark.parametrize('back', [True, False])
def test_compile(pp, back):
    Smartg(pp=pp, back=back)

def test_basic(request):
    ''' Most basic test '''
    m = Smartg(autoinit=True).run(500., atm=AtmAFGL('afglms'), NBPHOTONS=NBPHOTONS)
    smartg_view(m)
    conftest.savefig(request)

@pytest.mark.parametrize('wav', wav_list)
def test_atm(sg, wav):
    atm = AtmAFGL('afglt', comp=[AerOPAC('desert', 0.1, 550.)])
    m = sg.run(wav, atm=atm, NBPHOTONS=NBPHOTONS)
    assert ('wavelength' in m.axes) == ('__getitem__' in dir(wav))

@pytest.mark.parametrize('wav', wav_list)
def test_cloud(sg, wav):
    atm = AtmAFGL('afglt',
                  comp=[AerOPAC('desert', 0.1, 550.),
                        Cloud('wc', 12.68, 2, 3, 10., 550.),
                       ],
                  grid=[100., 50., 20., 10., 5., 2., 1., 0.],
                  pfgrid=[100., 10., 0.])
    m = sg.run(wav, atm=atm, NBPHOTONS=NBPHOTONS)
    assert ('wavelength' in m.axes) == ('__getitem__' in dir(wav))

@pytest.mark.parametrize('wav', wav_list)
@pytest.mark.parametrize('thv', [0., 40.])
@pytest.mark.parametrize('surf', [RoughSurface(WIND=2.), LambSurface(ALB=Albedo_cst(0.2))])
def test_atm_surf(sg, wav, surf, thv):
    atm = AtmAFGL('afglt', comp=[AerOPAC('desert', 0.1, 550.)])

    sg.run(wav,
           atm=atm,
           surf=surf,
           THVDEG=thv,
           NBPHOTONS=NBPHOTONS)


def test_surf_iop1_1():
    surf = RoughSurface(WIND=10.)
    water = IOP_1(chl=1.)
    Smartg().run([400., 500.], surf=surf, water=water, NBPHOTONS=NBPHOTONS)


def test_atm_surf_iop1():
    atm = AtmAFGL('afglt', comp=[AerOPAC('desert', 0.1, 550.)],
                  pfwav=[500., 600.], pfgrid=[100., 5., 0.])
    surf = RoughSurface(WIND=10.)
    water = IOP_1(chl=1., pfwav=np.array([450, 550, 650, 750]))
    wav = np.linspace(400, 800, 12)
    Smartg().run(wav, atm=atm, surf=surf, water=water, NBPHOTONS=NBPHOTONS)


def test_reptran(sg):
    atm = AtmAFGL('afglt', comp=[AerOPAC('desert', 0.1, 550.)])
    surf = RoughSurface(WIND=2.)

    ibands = REPTRAN('reptran_solar_msg').to_smartg('msg1')

    res = sg.run(ibands.l, atm=atm, surf=surf, water=None, NBPHOTONS=NBPHOTONS)
    reduce_reptran(res, ibands)


def test_locale_estimate(sg):
    atm = AtmAFGL('afglt')
    surf = RoughSurface()
    wav = 400.
    res = sg.run(wav, atm=atm,
                 surf=surf,
                 THVDEG=10.,
                 le={
                     'th_deg':  np.array([40.], dtype='float32'),
                     'phi_deg': np.array([30.], dtype='float32'),
                     },
                 NBPHOTONS=NBPHOTONS)
    assert res['I_up (TOA)'][:,:] > 0


@pytest.mark.parametrize('rng', ['PHILOX', 'CURAND_PHILOX'])
def test_rng(rng):
    atm = AtmAFGL('afglt')
    surf = RoughSurface()
    wav = np.linspace(400, 800, 5)
    Smartg(rng=rng).run(
        wav,
        atm=atm,
        surf=surf,
        NBPHOTONS=NBPHOTONS)


def test_adjacency():
    pytest.skip("Cannot test this, it is still in progress.")


def test_no_aer_output():
    atm1 = AtmAFGL('afglt')
    water = IOP_1(chl=0.5, DEPTH=5.)
    surf = RoughSurface(WIND=5., NH2O=1.34)
    sg = Smartg()
    le = {'th_deg': np.array([0., 45., 89.9]), 'phi_deg': np.array([0., 15.6, 289.])}
    m1 = sg.run(550., atm=atm1, surf=surf, water=water, OUTPUT_LAYERS=3, le=le,
                NBPHOTONS=1e6, NBLOOP=1e6, no_aer_output=True)
    m1 = m1.to_xarray() # to prepare transition luts to xarray
    assert (np.allclose(m1['I_up (TOA)'].values, m1['I_up (TOA), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['Q_up (TOA)'].values, m1['Q_up (TOA), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['U_up (TOA)'].values, m1['U_up (TOA), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['V_up (TOA)'].values, m1['V_up (TOA), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['I_down (0+)'].values, m1['I_down (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['Q_down (0+)'].values, m1['Q_down (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['U_down (0+)'].values, m1['U_down (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['V_down (0+)'].values, m1['V_down (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['I_down (0-)'].values, m1['I_down (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['Q_down (0-)'].values, m1['Q_down (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['U_down (0-)'].values, m1['U_down (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['V_down (0-)'].values, m1['V_down (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['I_up (0+)'].values, m1['I_up (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['Q_up (0+)'].values, m1['Q_up (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['U_up (0+)'].values, m1['U_up (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['V_up (0+)'].values, m1['V_up (0+), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['I_up (0-)'].values, m1['I_up (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['Q_up (0-)'].values, m1['Q_up (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['U_up (0-)'].values, m1['U_up (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['V_up (0-)'].values, m1['V_up (0-), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['I_down (B)'].values, m1['I_down (B), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['Q_down (B)'].values, m1['Q_down (B), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['U_down (B)'].values, m1['U_down (B), no_aer'].values, 0, 1e-13, equal_nan=True))
    assert (np.allclose(m1['V_down (B)'].values, m1['V_down (B), no_aer'].values, 0, 1e-13, equal_nan=True))
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-


from __future__ import print_function, division, absolute_import
import numpy as np
from luts.luts import LUT, MLUT
from smartg.atmosphere import od2k, BPlanck
from pathlib import Path
from scipy.integrate import quad, simpson
from smartg.config import DIR_AUXDATA
from scipy.interpolate import interp1d
import netCDF4
from smartg.tools.interp import interp2, interp3

dir_reptran = DIR_AUXDATA / 'reptran'

def reduce_reptran(mlut, ibands, use_solar=False, integrated=False, extern_weights=None):
    '''
    Compute the final spectral signal from mlut output of smart_g and
    REPTRAN_IBAND_LIST weights
    '''
    we, wb, ex, dl, norm, norm_dl = ibands.get_weights()
    res = MLUT()
    for l in mlut:
        for pref in ['I_','Q_','U_','V_','transmission','flux'] :
            if pref in l.desc:
                if extern_weights is not None:
                    tmp = l.desc
                    l = l*extern_weights 
                    l.desc = tmp
                if use_solar : lr = (l*we*ex*dl).reduce(np.sum,'wavelength',grouping=wb.data)
                else         : lr = (l*we*dl   ).reduce(np.sum,'wavelength',grouping=wb.data)
                if integrated: lr = lr/norm
                else         : lr = lr/norm_dl

                res.add_lut(lr, desc=l.desc)
    res.attrs = mlut.attrs
    return res


def Reptran_Emission(mlut, ibands):
    '''
    Return Thermal emission
    '''
    bsgroup = ibands.get_groups()
    kabs    = od2k(mlut, 'OD_abs_atm') * 1e-3 # m-1
    z       = -mlut.axis('z_atm') * 1e3 # m
    wmin = np.unique([ib.band.wmin for ib in ibands.l])
    wmax = np.unique([ib.band.wmax for ib in ibands.l])
    Avg_B  = np.zeros((len(wmin), len(z)))
    for i,(wmin,wmax) in enumerate(zip(wmin,wmax)):    
        for j,T in enumerate(mlut['T_atm'].data):
            lmin, lmax = wmin*1e-9, wmax*1e-9 # m
            dl         = wmax-wmin # nm
            Avg_B[i,j] = quad(BPlanck, lmin, lmax, args=T)[0]/(dl)
    Emission = LUT(kabs * Avg_B[bsgroup, :], 
               axes = [mlut.axis('wavelength'), z], 
               names= ['wavelength','z_atm'])
    return Emission


def Reptran_Avg_Emission(mlut, ibands):
    '''
    Return vertically integrated Thermal emission
    '''
    return (4*np.pi)*Reptran_Emission(mlut, ibands).reduce(simpson, 'z_atm', x=-mlut.axis('z_atm') * 1e3)



class REPTRAN_IBAND(object):
    '''
    REPTRAN internal band

    Arguments:
        band: REPTRAN_BAND object
        index: band index
        iband: internal band index
    '''
    def __init__(self, band, index):

        self.band = band     # parent REPTRAN_BAND
        self.index = index   # internal band index
        self.w = band.awvl[index]  # band wavelength
        self._iband=band._iband[index]
        self.weight =  band.awvl_weight[index]  # weight
        self.extra = band.aextra[index]  # solar irradiance
        self.crs_source = band.across_section_source[index,:]  # table of absorbing gases
        self.species=['H2O','CO2','O3','N2O','CO','CH4','O2','N2']
        self.filename = Path(band.filename)

    def calc_profile(self, prof):
        '''
        calculate a gaseous absorption profile for this internal band
        using temperature T and pressure P, and profile of molecular density of
        various gases stored in the profile prof
        '''
        Nmol = 8
        T = prof.T
        P = prof.P
        M = len(T)

        densmol = np.zeros((M, Nmol), np.float64)
        densmol[:,0] = prof.dens_h2o
        densmol[:,1] = prof.dens_co2
        densmol[:,2] = prof.dens_o3
        densmol[:,3] = prof.dens_no2
        densmol[:,4] = prof.dens_co
        densmol[:,5] = prof.dens_ch4
        densmol[:,6] = prof.dens_o2
        densmol[:,7] = prof.dens_n2

        xh2o = prof.dens_h2o/prof.dens_air

        datamol = np.zeros(M, np.float64)

        assert len(prof.T) == len(prof.P)

        # for each gas
        for ig in np.arange(Nmol):

            # si le gaz est absorbant a cette lambda
            if self.crs_source[ig]==1:

                # on recupere la LUT d'absorption
                crs_filename = self.filename.with_suffix('')  # supprime l'extension
                crs_filename = crs_filename.with_name(f"{crs_filename.name}.lookup.{self.species[ig]}")
                crs_mol = readCRS(crs_filename, self._iband)

                # interpolation du profil vertical de temperature de reference dans les LUT
                f = interp1d(crs_mol.pressure,crs_mol.t_ref, fill_value='extrapolate')
                #f = interp1d(crs_mol.pressure,crs_mol.t_ref)

                # ecart en temperature par rapport au profil de reference (ou P de reference est en Pa et P AFGL en hPa)
                dT = T - f(P*100)

                if ig == 0 :  # si h2o
                    # interpolation dans la LUT d'absorption en fonction de
                    # pression, ecart en temperature et vmr de h2o et mutiplication par la densite,
                    # calcul de reptran avec LUT en 10^(-20) m2, passage en km-1
                    datamol += interp3(crs_mol.t_pert,crs_mol.vmrs,crs_mol.pressure,crs_mol.xsec,dT,xh2o,P*100) * densmol[:,ig] * 1e-11
                else:
                    tab = crs_mol.xsec
                    # interpolation dans la LUT d'absorption en fonction de
                    # pression, ecart en temperature et mutiplication par la densite,
                    # calcul de reptran avec LUT en 10^(-20) m2, passage en km-1 
                    datamol += interp2(crs_mol.t_pert,crs_mol.pressure,np.squeeze(tab),dT,P*100) * densmol[:,ig] * 1e-11

        return datamol


class REPTRAN_BAND(object):
    def __init__(self, reptran, band):

        self.band = band
        self.nband = reptran.nwvl_in_band[self.band] # the number of internal bands (representative bands) in this channel
        self._iband = reptran.iwvl[:self.nband,self.band] # the indices of the internal bands within the wavelength grid for this channel
        self.awvl = reptran.wvl[self._iband-1] # the corresponsing wavelenghts of the internal bands
        self.awvl_weight = reptran.iwvl_weight[:self.nband,self.band] # the weights of the internal bands for this channel
        self.aextra = reptran.extra[self._iband-1] # the extra terrestrial solar irradiance of the internal bands for this channel
        self.across_section_source = reptran.cross_section_source[self._iband-1] # the source of absorption by species of the internal bands for this channel
        self.name = reptran.band_names[band]
        self.filename = Path(reptran.filename)    
        # the wavelength integral (width) of this channel
        self.Rint = reptran.wvl_integral[self.band]
        
        try:
            self.wmin = float(self.name.split('to')[0].rstrip().split('bandfrom')[1])
            self.wmax = float(self.name.split('to')[1].rstrip().split('nm')[0])
        except:
            self.w    = np.mean(self.awvl)
            self.wmin = self.w - self.Rint/2.
            self.wmax = self.w + self.Rint/2.


    def iband(self, index):
        '''
        returns internal band by its number (starting at zero)
        '''
        return REPTRAN_IBAND(self, index)

    def ibands(self):
        '''
        iterate over each internal band
        '''
        for i in range(self.nband):
            yield self.iband(i)
            

class REPTRAN(object):
    '''
    REPTRAN correlated-k file
    if provided without a directory, look to auxdata/reptran directory
    '''

    def __init__(self,filename):
        filename = Path(filename)
        if filename.parent == Path('.'):
            self.filename = dir_reptran / filename
        else:
            self.filename = filename

        if not filename.suffix == '.cdf':
            self.filename = self.filename.with_name(self.filename.name + '.cdf')

        self._readFileGeneral()

    def _readFileGeneral(self):
        nc = netCDF4.Dataset(self.filename)
        self.wvl = nc.variables['wvl'][:] # the wavelength grid
        if 'extra' in nc.variables.keys():
            self.extra = nc.variables['extra'][:] # the extra terrestrial solar irradiance for the walength grid
        else:
            self.extra = np.ones_like(self.wvl)
        self.wvl_integral = nc.variables['wvl_integral'][:] # the wavelength integral (width) of each sensor channel
        self.nwvl_in_band = nc.variables['nwvl_in_band'][:] # the number of internal bands (representative bands) in each sensor channel
        self.iwvl = nc.variables['iwvl'][:] # the indices of the internal bands within the wavelength grid for each sensor channel
        self.iwvl_weight = nc.variables['iwvl_weight'][:] # the weight associated to each internal band
        self.cross_section_source = nc.variables['cross_section_source'][:] # for each internal band, the list of species that participated to the absorption computation 

        self.band_names = []
        for bname in nc.variables['band_name']:  # the names of the sensor channels
            self.band_names.append(str(bname.tobytes()).replace(' ', ''))

    def nbands(self):
        '''
        number of bands
        '''
        return len(self.wvl_integral)

    def band(self, band):
        '''
        returns a REPTRAN_BAND
        band can be defined either by an integer, or a string
        '''
        if isinstance(band, str):
            return self.band(self.band_names.index(band))
        else:
            return REPTRAN_BAND(self, band)

    def bands(self):
        '''
        iterates over all bands
        '''
        for i in range(self.nbands()):
            yield self.band(i)

    def to_smartg(self, include='', lmin=-np.inf, lmax=np.inf,band_indices=None ):
        '''
        return a REPTRAN_IBAND_LIST for Smartg.run() method
        '''
        ik_l=[]
        if band_indices is None:
            bl = self.bands()
        else:
            bl = [self.band(i) for i in band_indices]
            
        if not isinstance(lmin,(list,np.ndarray)):
            lmin=[lmin]
            lmax=[lmax]
        for k in bl:
            if (include in k.name):
                for ii in range(len(lmin)):
                    if (k.wmin >= lmin[ii]) and (k.wmax <= lmax[ii]):
                        for ik in k.ibands():
                            ik_l.append(ik)

        assert len(ik_l) != 0

        return REPTRAN_IBAND_LIST(sorted(ik_l, key=lambda x:x.w))

class REPTRAN_IBAND_LIST(object):
    '''
    REPTRAN LIST OF IBANDS
    '''

    def __init__(self, l):
        self.l=l

    def get_weights(self):
        '''
        return weights, wavelengths, solarflux, bandwidth, bandwidth weighted normalization in postprocessing
        as MLUT objects

        Outputs:
        weights, wavelengths, solarflux, bandwidth, norm_bandwidth , norm

        '''
        we_l=[]
        ex_l=[]
        dl_l=[]
        wb_l=[]
        wi_l=[]
        for iband in self.l:
            #for iband in band.ibands():
                wi = iband.band.awvl[iband.index] # wvl of internal band
                wi_l.append(wi)
                we = iband.band.awvl_weight[iband.index] # weight of internal band
                we_l.append(we)
                ex = iband.band.aextra[iband.index] # E0 of internal band
                ex_l.append(ex)
                dl = iband.band.Rint # bandwidth
                dl_l.append(dl)
                wb = np.mean(iband.band.awvl[:])
                wb_l.append(wb)
        wb=LUT(np.array(wb_l),axes=[np.array(wi_l)],names=['wavelength'],desc='wavelength central band')
        we=LUT(np.array(we_l),axes=[np.array(wi_l)],names=['wavelength'],desc='Weight')
        ex=LUT(np.array(ex_l),axes=[np.array(wi_l)],names=['wavelength'],desc='E0')
        dl=LUT(np.array(dl_l),axes=[np.array(wi_l)],names=['wavelength'],desc='Dlambda')
        norm_dl = (we*dl).reduce(np.sum,'wavelength',grouping=wb.data)
        norm = we.reduce(np.sum,'wavelength',grouping=wb.data)
        return we, wb, ex, dl, norm, norm_dl 


    def get_groups(self):
        '''
        '''
        bsgroup=[]
        for iband in self.l:
            bsgroup.append(iband.band.band)
        bsgroup = np.array(bsgroup)
        return bsgroup-bsgroup[0]

        
    def get_names(self):
        '''
        return band names
        '''
        names=[]

        for iband in self.l:
            names.append(iband.band.name)

        return list(set(names))


class readCRS(object):
    def __init__(self,filename,iband):
        self.filename=Path(filename)
        self._readFileGeneral(iband)

    def _readFileGeneral(self,iband):
        nc=netCDF4.Dataset(dir_reptran / f'{self.filename.name}.cdf')
        self.wvl_index=nc.variables['wvl_index'][:]
        ii=list(self.wvl_index).index(iband)
        dat=nc.variables['xsec'][:]
        self.xsec=dat[:,:,ii,:]
        self.pressure=nc.variables['pressure'][:]
        self.t_ref=nc.variables['t_ref'][:]
        self.t_pert=nc.variables['t_pert'][:]
        self.vmrs=nc.variables['vmrs'][:]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division
from scipy.ndimage import map_coordinates
import numpy as np


def interp3(x, y, z, v, xi, yi, zi, **kwargs):
    """Sample a 3D array "v" with pixel corner locations at "x","y","z" at the
    points in "xi", "yi", "zi" using linear interpolation. Additional kwargs
    are passed on to ``scipy.ndimage.map_coordinates``."""
    assert v.ndim==3
    def index_coords(corner_locs, interp_locs):
        index = np.arange(len(corner_locs))
        if np.all(np.diff(corner_locs) < 0):
            corner_locs, index = corner_locs[::-1], index[::-1]
        return np.interp(interp_locs, corner_locs, index)

    orig_shape = np.asarray(xi).shape
    xi, yi, zi = np.atleast_1d(xi, yi, zi)
    for arr in [xi, yi, zi]:
        arr.shape = -1

    output = np.empty(xi.shape, dtype=float)
    coords = [index_coords(*item) for item in zip([x, y, z], [xi, yi, zi])]

    map_coordinates(v, coords, order=1, output=output, **kwargs)

    return output.reshape(orig_shape)

def interp2(x, y, v, xi, yi, **kwargs):
    """Sample a 2D array "v" with pixel corner locations at "x","y", at the
    points in "xi", "yi",  using linear interpolation. Additional kwargs
    are passed on to ``scipy.ndimage.map_coordinates``."""
    assert v.ndim==2
    def index_coords(corner_locs, interp_locs):
        index = np.arange(len(corner_locs))
        if np.all(np.diff(corner_locs) < 0):
            corner_locs, index = corner_locs[::-1], index[::-1]
        return np.interp(interp_locs, corner_locs, index)

    orig_shape = np.asarray(xi).shape
    xi, yi = np.atleast_1d(xi, yi)
    for arr in [xi, yi]:
        arr.shape = -1

    output = np.empty(xi.shape, dtype=float)
    coords = [index_coords(*item) for item in zip([x, y], [xi, yi])]

    map_coordinates(v, coords, order=1, output=output, **kwargs)

    return output.reshape(orig_shape)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pycuda.driver as drv
import pycuda.autoinit
import numpy
from pycuda.compiler import SourceModule
from pycuda.gpuarray import zeros as gpuzeros


def test_pycuda():
    '''
    A basic pycuda test, from pycuda examples hello_gpu
    '''
    mod = SourceModule("""
    __global__ void multiply_them(float *dest, float *a, float *b)
    {
    const int i = threadIdx.x;
    dest[i] = a[i] * b[i];
    }
    """)

    multiply_them = mod.get_function("multiply_them")

    a = numpy.random.randn(400).astype(numpy.float32)
    b = numpy.random.randn(400).astype(numpy.float32)

    dest = numpy.zeros_like(a)
    multiply_them(
            drv.Out(dest), drv.In(a), drv.In(b),
            block=(400, 1, 1))

    numpy.testing.assert_allclose(dest-a*b, 0)
    print('Used', pycuda.autoinit.device.name())


def test_atomic_add():
    mod = SourceModule("""
    __global__ void mykernel(int *a)
    {
        atomicAdd(a, 1);
    }
    """)
    mykernel = mod.get_function("mykernel")
    a = gpuzeros(1, dtype='int32')
    mykernel(a, block=(256, 1, 1), grid=(256, 1, 1))
    assert a == 256*256
    print('Used', a, pycuda.autoinit.device.name())



#!/usr/bin/env python
# vim:fileencoding=utf-8

'''
Calculate the ocean water absorption, scattering coefficients and phase
function using the wavelength and the SPM concentration
'''

from numpy import sin, cos, pi, exp
from numpy import arange, zeros, log10, array
from phase_functions import fournierForand, PhaseFunction, fournierForandB
from iop import IOP
import sys
sys.path.append('/home/did/RTC/SMART-G/tools/water/cython')
from waterPR import ParkRuddick

rep='/home/francois/MERIS/polymer_py/auxdata/common/'


class IOP_PandR(IOP):
    '''
    Initialize the IOP model (Chl dominated waters)

    Arguments:
        CHL: Chlorophyll mg/m3
        NANG: number of angles for Phase function
        ang_trunc : truncature angle for Fournier Forand Phase function
        ALB: albedo of the sea floor
        DEPTH: depth in m , default None = Semi infinite
        pfwav: list of arrays at which the phase functions are calculated
    '''
    def __init__(self, CHL, NANG=72001, ang_trunc=5., ALB=0., DEPTH=None, pfwav=None, verbose=False):
        self.__CHL = CHL
        self.__NANG = NANG
        self.__ang_trunc = ang_trunc
        self.alb = ALB
        self.depth = DEPTH
        self.pfwav=pfwav
        self.verbose = verbose
        self.last = None   # stores last result (managed by parent class)


    def calc(self, w, skip_phase=False, nbp=1.15):
        '''
        Calculate atot, btot and phase function (monochromatic)

        Arguments:
            w: wavelength in nm

        Returns (atot, btot, phase) where:
            * atot is the total absorption coefficient
            * btot is the total scattering coefficient
            * phase is the PhaseFunction object
        '''

        PR = ParkRuddick(rep, debug=True, alt_gamma_bb=True)
        CHL = self.__CHL
        NANG = self.__NANG
        ang_trunc = self.__ang_trunc
        PR.calc(w, log10(CHL))
        self.iops = PR.iops()

        # pure sea water scattering 
        bw = self.iops['bw']

        # Scattering coefficient of particles
        btot = self.iops['btot']
        bp = btot - bw

        #
        # total absorption
        atot = self.iops['atot']

        if skip_phase:
            P0, P1 = None, None
        else:
            #
            # phase function
            #
            ang = pi * arange(NANG, dtype='float64')/(NANG-1)    # angle in radians

            # pure water
            pf0 = zeros((NANG, 4), dtype='float64') # pure water phase function
            pf0[:,1] = 0.75
            pf0[:,0] = 0.75 * cos(ang)**2
            pf0[:,2] = 0.75 * cos(ang)
            pf0[:,3] = 0.
            P0 = PhaseFunction(ang, 2*pf0, degrees=False)

            # particles (troncature)
            # particle slope of the scattering spectral dependency
            gamma = self.iops['gamma']
            itronc = int(NANG * ang_trunc/180.)
            pf1 = zeros((NANG, 4), dtype='float64') # pure water phase function
            # assuming that the slope of Junge power law mu and slope of spectral dependence of scattering is mu=3+gamma
            pf1[itronc:,0] = 0.5*fournierForand(ang[itronc:],nbp,3.+gamma)
            pf1[:itronc,0] = 0.5*fournierForand(ang[itronc ],nbp,3.+gamma) 
            pf1[:,1] = pf1[:,0]
            pf1[:,2] = 0.
            pf1[:,3] = 0.

            # normalization after truncation
            integ_ff = 0.
            integ_ff_back = 0.
            for iang in xrange(1, NANG):
                dtheta = ang[iang] - ang[iang-1]
                pm1 = pf1[iang-1,0] + pf1[iang-1,1]
                pm2 = pf1[iang,0] + pf1[iang,1]
                sin1 = sin(ang[iang-1])
                sin2 = sin(ang[iang])
                integ_ff += dtheta*((sin1*pm1+sin2*pm2)/3. + (sin1*pm2+sin2*pm1)/6.)
                if ang[iang]>pi/2. :
                    integ_ff_back += dtheta*((sin1*pm1+sin2*pm2)/3. + (sin1*pm2+sin2*pm1)/6.)
            rat1 = integ_ff/2.
            pf1 *= 1/rat1
            P1 = PhaseFunction(ang, 2*pf1, degrees=False, coef_trunc=rat1)

        return atot, [(bw, P0), (bp, P1)]

    def __str__(self):
        return 'SPM={}'.format(self.__SPM)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import
import numpy as np
from numpy.core.fromnumeric import shape
from luts.luts import LUT, MLUT
from smartg.atmosphere import od2k, BPlanck
from scipy.integrate import quad, simpson
import sys
from scipy.interpolate import interp1d
from scipy.interpolate import interpn
import os
from itertools import product
from smartg.tools.interp import interp2
import h5py

from smartg.config import DIR_AUXDATA
from pathlib import Path
dir_kdis = DIR_AUXDATA / 'kdis'
import warnings

import glob

def reduce_kdis(mlut, ibands, use_solar=False, integrated=False, extern_weights=None):
    '''
    Compute the final spectral signal from mlut output of smart_g and
    KDIS_IBAND_LIST weights
    '''
    we, wb, ex, dl, norm, norm_dl = ibands.get_weights()
    res = MLUT()
    for l in mlut:
        for pref in ['I_','Q_','U_','V_','N_','transmission','flux'] :
             if pref in l.desc:
                if extern_weights is not None:
                    tmp = l.desc
                    l = l*extern_weights 
                    l.desc = tmp
                if use_solar : lr = (l*we*ex*dl).reduce(np.sum,'wavelength',grouping=wb.data)
                else         : lr = (l*we*dl   ).reduce(np.sum,'wavelength',grouping=wb.data)
                if integrated: lr = lr/norm
                else         : lr = lr/norm_dl
                res.add_lut(lr, desc=l.desc)
    res.attrs = mlut.attrs
    return res



def Kdis_Emission(mlut, ibands):
    '''
    Return Thermal emission
    '''
    bsgroup = ibands.get_groups()
    kabs    = od2k(mlut, 'OD_abs_atm') * 1e-3 # m-1
    z       = -mlut.axis('z_atm') * 1e3 # m
    wmin = np.unique([ib.band.wmin for ib in ibands.l])
    wmax = np.unique([ib.band.wmax for ib in ibands.l])
    Avg_B  = np.zeros((len(wmin), len(z)))
    for i,(wmin,wmax) in enumerate(zip(wmin,wmax)):    
        for j,T in enumerate(mlut['T_atm'].data):
            lmin, lmax = wmin*1e-9, wmax*1e-9 # m
            dl         = wmax-wmin # nm
            Avg_B[i,j] = quad(BPlanck, lmin, lmax, args=T)[0]/(dl)
    Emission = LUT(kabs * Avg_B[bsgroup, :], 
               axes = [mlut.axis('wavelength'), z], 
               names= ['wavelength','z_atm'])

    return Emission


def Kdis_Avg_Emission(mlut, ibands):
    '''
    Return vertically integrated Thermal emission
    '''
    return (4*np.pi)*Kdis_Emission(mlut, ibands).reduce(simpson, 'z_atm', x=-mlut.axis('z_atm') * 1e3)


class KDIS(object):

    def __init__(self, model, dir_data='', format=None):

        # read the entire K-distribution definition from files
        #
        # Selection of the desired KDIS band or absorbing gases
        # must be done later while setting up the artdeco variables
        # if dir_data is not specified, standard dir is assumed to be auxdata/kdis dir

        self.model = model

        dir_data = Path(dir_data)
        if dir_data.parent == Path('.') : dir_data = dir_kdis / dir_kdis / model
    
        is_sorted = lambda a: np.all(a[:-1] <= a[1:])

        if format is None:
            if (len(glob.glob(str(dir_data) + '/*.h5')) > 0) : format = 'h5'
            else                                             : format = 'ascii'
        else:
            warnings.simplefilter('always', DeprecationWarning)
            warn_message = "\nThe key argument 'format' is now useless and deprecated as of SMART-G 1.0.0,\n" + \
                        "and will be removed in one of the next release."
            warnings.warn(warn_message, DeprecationWarning)
 
        if format == 'ascii':
            filename = dir_data / f'kdis_{model}_def.dat'
            if not filename.is_file():
                print("(kdis_coef) ERROR")
                print("            Missing file:", filename)
                sys.exit()
            fdef = open(filename,'r')
            skipcomment(fdef)
            tmp = fdef.readline()
            self.nmaxai = int(tmp.split()[0])
            skipcomment(fdef)
            tmp = fdef.readline()
            self.nsp_tot =  int(tmp.split()[0])
            self.nsp     = 0
            self.fcont   = []
            self.species = []
            self.nsp_c     = 0
            self.fcont_c   = []
            self.species_c = []
            skipcomment(fdef)
            for i in range(self.nsp_tot):
                tmp = fdef.readline()
                if int(tmp.split()[1]) == 0:
                    self.nsp  = self.nsp  + 1
                    self.species.append(tmp.split()[0])
                    self.fcont.append( float(tmp.split()[2] ) )
                elif  int(tmp.split()[1]) == 1:
                    self.nsp_c  = self.nsp_c  + 1
                    self.species_c.append(tmp.split()[0])
                    self.fcont_c.append( float(tmp.split()[2] ) )
            self.fcont   = np.array(self.fcont)        
            self.fcont_c = np.array(self.fcont_c)        
            skipcomment(fdef)
            tmp = fdef.readline()
            self.nwvl = int(tmp.split()[0])
            self.wvlband = np.zeros((3, self.nwvl))
            skipcomment(fdef)
            for i in range(self.nwvl):
                tmp = fdef.readline()                
                self.wvlband[0,i] = float(tmp.split()[1])*1e3
                self.wvlband[1,i] = float(tmp.split()[2])*1e3
                self.wvlband[2,i] = float(tmp.split()[3])*1e3
                if i>0:
                    if self.wvlband[0,i] < self.wvlband[0,i-1]:
                        print(" kdis_coeff ERROR")
                        print("            wavelengths must be sorted in increasing order")
                        sys.exit()
            skipcomment(fdef)
            tmp = fdef.readline()
            skipcomment(fdef)
            self.np = int(tmp.split()[0])
            self.p = np.zeros(self.np)
            for i in range(self.np):
                tmp = fdef.readline()                
                self.p[i] = float(tmp.split()[0])
                if i>0:
                    if self.p[i] < self.p[i-1]:
                        print(" kdis_coeff ERROR")
                        print("            pressure must be sorted in increasing order")
                        sys.exit()
            skipcomment(fdef)
            tmp = fdef.readline()
            skipcomment(fdef)
            self.nt = int(tmp.split()[0])
            self.t = np.zeros(self.nt)
            for i in range(self.nt):
                tmp = fdef.readline()                
                self.t[i] = float(tmp.split()[0])
                if i>0:
                    if self.t[i] < self.t[i-1]:
                        print(" kdis_coeff ERROR")
                        print("            temperature must be sorted in increasing order")
                        sys.exit()
            if self.nsp_c > 0:
                skipcomment(fdef)
                tmp = fdef.readline()
                skipcomment(fdef)
                self.nc = int(tmp.split()[0])
                self.c = np.zeros(self.nc)
                for i in range(self.nc):
                    tmp = fdef.readline()                
                    self.c[i] = float(tmp.split()[0])
                    if i>0:
                        if self.c[i] < self.c[i-1]:
                            print(" kdis_coeff ERROR")
                            print("            concentration must be sorted in increasing order")
                            sys.exit()
            fdef.close()
            if self.nsp > 0:
                self.nai   = np.zeros((self.nsp,self.nwvl), dtype='int')
                self.ki    = np.zeros((self.nsp,self.nwvl,self.nmaxai,self.np,self.nt))
                self.ai    = np.zeros((self.nsp,self.nwvl,self.nmaxai))
            if self.nsp_c > 0:
                self.nai_c   = np.zeros((self.nsp_c,self.nwvl), dtype='int')
                self.ki_c    = np.zeros((self.nsp_c,self.nwvl,self.nmaxai,self.np,self.nt,self.nc))
                self.ai_c    = np.zeros((self.nsp_c,self.nwvl,self.nmaxai))
            for isp in range(self.nsp):
                filename = dir_data / f'kdis_{model}_{self.species[isp]}.dat'
                if not filename.is_file():
                    print("(kdis_coef) ERROR")
                    print("            Missing file:", filename)
                    sys.exit()                    
                f = open(filename,'r')
                skipcomment(f)
                for iwvl in range(self.nwvl):
                    tmp = f.readline()
                    self.nai[isp,iwvl]   = int(tmp.split()[1])
                for iwvl in range(self.nwvl):
                    if self.nai[isp,iwvl]>1:
                        skipcomment(f)
                        tmp = f.readline()
                        #print 'nai, nmaxai=',self.nai[isp,iwvl], self.nmaxai
                        for iai in range(self.nai[isp,iwvl]):
                            #print iai, float(tmp.split()[iai])  
                            self.ai[isp,iwvl,iai] = float(tmp.split()[iai])  
                        for it in range(self.nt):
                            for ip in range(self.np):
                                tmp = f.readline()
                                for iai in range(self.nai[isp,iwvl]):
                                    self.ki[isp,iwvl,iai,ip,it] = float(tmp.split()[iai])                                  
                f.close()
            if self.nsp_c > 0:
                self.c_desc = "density"
            else:
                self.c_desc = "none"
            for isp in range(self.nsp_c):
                filename = dir_data / f'kdis_{model}_{self.species_c[isp]}.dat'
                if not filename.is_file():
                    print("(kdis_coef) ERROR")
                    print("            Missing file:", filename)
                    sys.exit()                    
                f = open(filename,'r')
                skipcomment(f)
                for iwvl in range(self.nwvl):
                    tmp = f.readline()
                    self.nai_c[isp,iwvl]   = int(tmp.split()[1])
                for iwvl in range(self.nwvl):
                    if self.nai_c[isp,iwvl]>1:
                        skipcomment(f)
                        tmp = f.readline()
                        for iai in range(self.nai_c[isp,iwvl]):
                            self.ai_c[isp,iwvl,iai] = float(tmp.split()[iai])  
                        for ic in range(self.nc):
                            for it in range(self.nt):
                                for ip in range(self.np):
                                    tmp = f.readline()
                                    for iai in range(self.nai_c[isp,iwvl]):
                                        self.ki_c[isp,iwvl,iai,ip,it,ic] = float(tmp.split()[iai])                                  
                f.close()
            
            filename = dir_data / f'kdis_{model}_solarflux.dat'
            if not filename.is_file():
                filename = dir_data /f'solrad_kdis_{model}_thuillier2003.dat'
                if not filename.is_file():
                    print("(kdis_coef) ERROR")
                    print("            Missing file:", filename)
                    sys.exit()
            fsol = open(filename,'r')
            skipcomment(fsol)
            tmp = fsol.readline()
            skipcomment(fsol)
            tmp = fsol.readline()
            nn = float(tmp.split()[0])
            if nn != self.nwvl :
                print(" solar flux and kdis have uncompatible band number")
                sys.exit()
            skipcomment(fsol)
            self.solarflux = np.zeros(self.nwvl)
            skipcomment(fsol)
            for i in range(self.nwvl):
                tmp = fsol.readline()
                self.solarflux[i] = float(tmp.split()[0])
            fsol.close()
        
        elif format in ["h5","hdf5"]:

            filename = dir_data / f'kdis_{model}.h5'
            f = h5py.File(filename,"r")
            self.nmaxai = np.copy(f["def"]["maxnai"])
            species_tot = list(f["coeff"].keys())
            self.nsp_tot =  len(species_tot)
            self.nsp     = 0
            self.fcont   = []
            self.species = []
            self.nsp_c     = 0
            self.fcont_c   = []
            self.species_c = []
            for isp, specie in enumerate(species_tot):
                if "rho_dep" in list(f["coeff"][specie].attrs.keys()):
                    if f["coeff"][specie].attrs['rho_dep']:
                        rho_dep = True
                    else:
                        rho_dep = False
                else:
                    rho_dep = False
                if not rho_dep:
                    self.nsp  = self.nsp  + 1
                    self.species.append(specie)
                    self.fcont.append( f["coeff"][specie].attrs['add_continuum'] )
                else:
                    self.nsp_c  = self.nsp_c  + 1
                    self.species_c.append(specie)
                    self.fcont_c.append(  f["coeff"][specie].attrs['add_continuum'] )
            self.fcont   = np.array(self.fcont)        
            self.fcont_c = np.array(self.fcont_c)        
            self.nwvl = len(f['def']['central_wvl'])
            self.wvlband = np.zeros((3, self.nwvl))
            self.wvlband[0,:] = np.copy(f['def']['central_wvl'])*1e3
            self.wvlband[1,:] = np.copy(f['def']['min_wvl'])*1e3
            self.wvlband[2,:] = np.copy(f['def']['max_wvl'])*1e3
            self.p = np.copy(f['def']['pressure'])
            self.t = np.copy(f['def']['temperature'])
            self.np = len(self.p)
            self.nt = len(self.t)            
            if self.nsp_c > 0:
                self.c = np.copy(f['def']['rho'])
                self.c_desc = f['def']['rho'].attrs["desc"].decode()
                self.nc = len(self.c)
                if not is_sorted(self.c):
                    print(" kdis_coeff ERROR")
                    print("            concentration must be sorted in increasing order")
                    sys.exit()
            else:
                self.c_desc = "none"
            if not is_sorted(self.wvlband[0,:]):
                print(" kdis_coeff ERROR")
                print("            (h5 format) read NOT implemented for concentration dependent species")
                sys.exit()
            if not is_sorted(self.p):
                print(" kdis_coeff ERROR")
                print("            pressure must be sorted in increasing order")
                sys.exit()
            if not is_sorted(self.t):
                print(" kdis_coeff ERROR")
                print("            temperature must be sorted in increasing order")
                sys.exit()
            if self.nsp>0:
                self.nai   = np.zeros((self.nsp,self.nwvl), dtype='int')
                self.ki    = np.zeros((self.nsp,self.nwvl,self.nmaxai,self.np,self.nt))
                self.ai    = np.zeros((self.nsp,self.nwvl,self.nmaxai))
                for isp, specie in enumerate(self.species):
                    self.nai[isp,:]      = np.copy(f["coeff"][specie]["nai"])
                    nai_tmp = np.nanmax(self.nai[isp,:])    
                    self.ki[isp,:,0:nai_tmp,:,:] = np.copy(f["coeff"][specie]["ki"][:,0:nai_tmp,:,:]) 
                    self.ai[isp,:,0:nai_tmp]     = np.copy(f["coeff"][specie]["ai"][:,0:nai_tmp]) 
            if self.nsp_c>0:
                self.nai_c   = np.zeros((self.nsp_c,self.nwvl), dtype='int')
                self.ki_c    = np.zeros((self.nsp_c,self.nwvl,self.nmaxai,self.np,self.nt,self.nc))
                self.ai_c    = np.zeros((self.nsp_c,self.nwvl,self.nmaxai))
                for isp, specie in enumerate(self.species_c):
                    self.nai_c[isp,:]      = np.copy(f["coeff"][specie]["nai"])
                    nai_tmp = np.nanmax(self.nai_c[isp,:])    
                    self.ki_c[isp,:,0:nai_tmp,:,:] = np.copy(f["coeff"][specie]["ki"][:,0:nai_tmp,:,:]) 
                    self.ai_c[isp,:,0:nai_tmp]     = np.copy(f["coeff"][specie]["ai"][:,0:nai_tmp]) 
            #solar flux
            grp  = f.require_group("solrad")
            self.solarflux = grp["solrad"][:]            
            f.close()

            for isp, specie in enumerate(self.species):
                self.species[isp] = self.species[isp].lower() 
            for isp, specie in enumerate(self.species_c):
                self.species_c[isp] = self.species_c[isp].lower()

        # support for multi species   
        if  (self.nsp>0) and (self.nsp_c>0) : 
            self.nai_eff    = np.prod( np.append(self.nai, self.nai_c, axis=0 ) , axis=0 )
        elif self.nsp>0 :
            self.nai_eff    = np.prod( self.nai , axis=0 )
        elif self.nsp_c>0 :
            self.nai_eff    = np.prod( self.nai_c, axis=0 )
        self.nmaxai_eff = np.max(self.nai_eff)        
        self.ai_eff     = np.zeros((self.nwvl,self.nmaxai_eff))
        if  self.nsp>0:
            self.iki_eff    = np.zeros((self.nsp,  self.nwvl,self.nmaxai_eff), dtype='int')
        if  self.nsp_c>0:    
            self.iki_eff_c  = np.zeros((self.nsp_c,self.nwvl,self.nmaxai_eff), dtype='int')
        for iwvl in range(self.nwvl):
            iai_eff = 0
            nested_list = []
            for isp in range(self.nsp):
                nested_list.append(range(self.nai[isp,iwvl]))
            for isp_c in range(self.nsp_c):
                nested_list.append(range(self.nai_c[isp_c,iwvl]))
            for k in product(*nested_list):
                if self.nsp>0:
                    self.iki_eff[:,iwvl, iai_eff]   = k[0:self.nsp]
                if self.nsp_c>0:
                    self.iki_eff_c[:,iwvl, iai_eff] = k[self.nsp:self.nsp+self.nsp_c]
                self.ai_eff[iwvl, iai_eff]    = 1.0                
                for isp in range(self.nsp):
                    if (self.nai[isp,iwvl] >= 1) and (self.ai[isp,iwvl,self.iki_eff[isp,iwvl,iai_eff]] != 0.0):
                        self.ai_eff[iwvl, iai_eff ] *= self.ai[isp,iwvl,self.iki_eff[isp,iwvl,iai_eff]]    
                for isp_c in range(self.nsp_c):
                    if (self.nai_c[isp_c,iwvl] >= 1) and (self.ai_c[isp_c,iwvl,self.iki_eff_c[isp_c,iwvl,iai_eff]]!= 0.0):
                        self.ai_eff[iwvl, iai_eff ] *= self.ai_c[isp_c,iwvl,self.iki_eff_c[isp_c,iwvl,iai_eff]]
                iai_eff += 1

                
    def nbands(self):
        '''
        number of bands
        '''
        return self.nwvl

    def band(self, band):
        '''
        returns a KDIS_BAND
        '''
        return KDIS_BAND(self, band)

    def bands(self):
        '''
        iterates over all bands
        '''
        for i in range(self.nbands()):
            yield self.band(i)    
            

    def to_smartg(self, include='', lmin=-np.inf, lmax=np.inf, band_indices=None):
        '''
        return a list of KDIS_IBANDS for Smartg.run() method
        '''
        ik_l=[]
        if band_indices is None:
            bl = self.bands()
        else:
            bl = [self.band(i) for i in band_indices]

        if not isinstance(lmin,(list,np.ndarray)):
            lmin=[lmin]
            lmax=[lmax]
        for k in bl:
            for ii in range(len(lmin)):
                if (k.wmin >= lmin[ii]) and (k.wmax <= lmax[ii]):
                    for ik in k.ibands():
                        ik_l.append(ik)

        assert len(ik_l) != 0

        return KDIS_IBAND_LIST(sorted(ik_l, key=lambda x:x.w))


    def get_weight(self):
        '''
        return weights, wavelengths, solarflux, band width and normalization in postprocessing
        '''
        wb_l=[]
        we_l=[]
        ex_l=[]
        dl_l=[]
        for k in self.bands():
            for ik in k.ibands():
                wb_l.append(ik.w)
                we_l.append(ik.weight)
                ex_l.append(ik.ex)
                dl_l.append(ik.dl)
        wb=LUT(np.array(wb_l),axes=[wb_l],names=['wavelength'],desc='wavelength')
        we=LUT(np.array(we_l),axes=[wb_l],names=['wavelength'],desc='weight')
        ex=LUT(np.array(ex_l),axes=[wb_l],names=['wavelength'],desc='solarflux')
        dl=LUT(np.array(dl_l),axes=[wb_l],names=['wavelength'],desc='bandwidth')
        norm = we.reduce(np.sum,'wavelength',grouping=wb.data)
        return we, wb, ex, dl, norm
   
class KDIS_IBAND(object):
    '''
    KDIS internal band

    Arguments:
        band: KDIS_BAND object
        index: band index
        iband: internal band index
    '''
    def __init__(self, band, index):

        self.band = band     # parent KDIS_BAND
        self.index = index   # internal band index
        self.w = band.awvl[index]  # band wavelength
        self.ex= band.solarflux # solar irradiance
        self.dl= band.dl  #bandwidth
#        if self.xsect != 0 : self.weight =  band.awvl_weight[index]  # weight
#        else : self.weight = 1.
        self.weight =  band.awvl_weight[index]  # weight
        #self.species=['H2O','CO2','O3','N2O','CO','CH4','O2','N2']
           
                
    # def ki_interp(P, T, ki, Pout, Tout):
    #     ki_t = np.zeros(len(T))
    #     for it in range(len(T)):
    #         fki        = interp1d(P, ki[:,it], kind='linear')
    #         ki_t[it] = fki(Pout)
    #     fki = interp1d(T, ki_t, kind='linear')
    #     return fki(Tout)
        

    def calc_profile(self, prof):
        '''
        calculate a gaseous absorption profile for this internal band
        using temperature T and pressure P, and profile of molecular density of
        various gases stored in densmol
        '''

        species = ['h2o', 'co2', 'o3', 'no2', 'co', 'ch4', 'o2', 'n2', 'n2o', 'so2']
        #species = ['h2o', 'co2', 'o3', 'n2o', 'co', 'ch4', 'o2', 'n2']
        T = prof.T.copy()
        P = prof.P.copy()
        Ngas = 10
        #Ngas = 8
        M = len(T)
        datamol = np.zeros(M, np.float64)

        densmol = np.zeros((M, Ngas), np.float64)
        densmol[:,0] = prof.dens_h2o[:]
        densmol[:,1] = prof.dens_co2[:]
        densmol[:,2] = prof.dens_o3[:]
        densmol[:,3] = prof.dens_no2[:]
        densmol[:,4] = prof.dens_co[:]
        densmol[:,5] = prof.dens_ch4[:]
        densmol[:,6] = prof.dens_o2[:]
        densmol[:,7] = prof.dens_n2[:]
        densmol[:,8] = prof.dens_n2o[:]
        densmol[:,9] = prof.dens_so2[:]

    
        # for each gas
        for ig_c in range(self.band.kdis.nsp_c):
            specie_c = self.band.kdis.species_c[ig_c]
            ispecie_c= species.index(specie_c)
            ikig_c = self.band.kdis.iki_eff_c[ig_c, self.band.band, self.index]
            tab = self.band.kdis.ki_c[ig_c, self.band.band, ikig_c, :, :, :]
            points = ( self.band.kdis.p, self.band.kdis.t, self.band.kdis.c )
            if self.band.kdis.c_desc == "density":
                C = densmol[:,ispecie_c] 
            elif self.band.kdis.c_desc == "molar_fraction":
                C = densmol[:,ispecie_c] / prof.dens_air.copy()
            C[C>np.max(self.band.kdis.c)]=np.max(self.band.kdis.c)*0.99
            C[C<np.min(self.band.kdis.c)]=np.min(self.band.kdis.c)*1.01            
            P[P>np.max(self.band.kdis.p)]=np.max(self.band.kdis.p)*0.99
            P[P<np.min(self.band.kdis.p)]=np.min(self.band.kdis.p)*1.01            
            T[T>np.max(self.band.kdis.t)]=np.max(self.band.kdis.t)*0.99
            T[T<np.min(self.band.kdis.t)]=np.min(self.band.kdis.t)*1.01            
            values = np.concatenate( (np.array([P]), np.array([T]), np.array([C])), axis=0).T
            #print(values)
            #print(interpn(points, tab, values))
            datamol +=  interpn(points, tab, values) * densmol[:,ispecie_c]

        for ig in range(self.band.kdis.nsp):
            specie = self.band.kdis.species[ig]
            ispecie= species.index(specie)
            ikig = self.band.kdis.iki_eff[ig, self.band.band, self.index]
            tab = self.band.kdis.ki[ig, self.band.band, ikig, :, :]
            datamol += interp2(self.band.kdis.p, self.band.kdis.t, np.squeeze(tab), P, T) * densmol[:,ispecie]

        return datamol*1e5


class KDIS_BAND(object):
    def __init__(self, kdis, band):

        self.kdis = kdis # parent kdis coeff
        self.band = band
        self.w = kdis.wvlband[0, self.band]
        self.wmin = kdis.wvlband[1, self.band]
        self.wmax = kdis.wvlband[2, self.band]
        #self.nband = kdis.nai[0, self.band] # the number of internal bands (representative bands) in this channel
        self.nband = kdis.nai_eff[self.band] # the number of internal bands (representative bands) in this channel
        #self.iband= np.arange(self.nband)
        self.awvl = [self.w]*self.nband # the corresponsing wavelenghts of the internal bands
        self.awvl_weight = kdis.ai_eff[self.band, :self.nband] # the weights of the internal bands for this channel
        #self.awvl_weight = kdis.ai[0, self.band, :self.nband] # the weights of the internal bands for this channel
        self.dl = (kdis.wvlband[2,self.band] - kdis.wvlband[1,self.band]) # bandwidth
        self.solarflux = kdis.solarflux[self.band]/self.dl # the extra terrestrial solar irradiance of the internal bands for this channel    


    def iband(self, index):
        '''
        returns internal band by its number (starting at zero)
        '''
        return KDIS_IBAND(self, index)

    def ibands(self):
        '''
        iterate over each internal band
        '''
        for i in range(self.nband):
            yield self.iband(i)

class KDIS_IBAND_LIST(object):  
    '''
    KDIS list of IBANDS
    '''
    def __init__(self, l):
        self.l=l

    def get_weights(self):
        '''
        return weights, wavelengths, solarflux, band width and normalization in postprocessing
        '''
        wb_l=[]
        we_l=[]
        ex_l=[]
        dl_l=[]
        for ik in self.l:
                wb_l.append(ik.w)
                we_l.append(ik.weight)
                ex_l.append(ik.ex)
                dl_l.append(ik.dl)
        wb=LUT(np.array(wb_l),axes=[wb_l],names=['wavelength'],desc='wavelength')
        we=LUT(np.array(we_l),axes=[wb_l],names=['wavelength'],desc='weight')
        ex=LUT(np.array(ex_l),axes=[wb_l],names=['wavelength'],desc='solarflux')
        dl=LUT(np.array(dl_l),axes=[wb_l],names=['wavelength'],desc='bandwidth')
        norm_dl = (we*dl).reduce(np.sum,'wavelength',grouping=wb.data)
        norm = we.reduce(np.sum,'wavelength',grouping=wb.data)
        return we, wb, ex, dl, norm, norm_dl    



    def get_groups(self):
        '''
        '''
        bsgroup=[]
        for iband in self.l:
            bsgroup.append(iband.band.band)
        bsgroup = np.array(bsgroup)
        return bsgroup-bsgroup[0]



def skipcomment(f):
    while(True):
        pos=f.tell()
        if not f.readline().strip().startswith('#'): break
    f.seek(pos,0)    



import scipy.constants as cst
import numpy as np

# Atmosphere model: mixing ratio (mol/mol)
X_N2 = 0.788
X_O2 = 0.212

def is_odd(j):
    return j % 2 != 0

# Bates, Planel. Space Sa., Vol.32, No.6, pp. 785-790. 1984 
def Fk_N2(lam):
    '''
    lam in nm
    '''
    return 1.034 + 3.17*1e-4/((lam*1e-3)**2)

def Epsilon_N2(lam):
    '''
    lam in nm
    '''
    return (Fk_N2(lam)-1) * 4.5

def Fk_O2(lam):
    '''
    lam in nm
    '''
    return 1.096 + 1.385*1e-3/((lam*1e-3)**2) + 1.448*1e-4/((lam*1e-3)**4)

def Epsilon_O2(lam):
    '''
    lam in nm
    '''
    return (Fk_O2(lam)-1) * 4.5

def Epsilon_air(lam):
    '''
    lam in nm
    '''
    return Epsilon_N2(lam) * X_N2 + Epsilon_O2(lam) * X_O2

#Kattawar, Astrophysical Journal, Part 1, vol. 243, Feb. 1, 1981, p. 1049-1057.
def f0_air(lam, theta):
    '''
    lam in nm
    theta in deg
    '''
    eps = Epsilon_air(lam)
    c2  = np.cos(np.radians(theta))**2
    num = (180.+13.*eps) + (180.+eps)*c2
    den = (180.+52.*eps) + (180.+4.*eps)*c2
    return num/den
    
def f0_N2(lam, theta):
    '''
    lam in nm
    theta in deg
    '''
    eps = Epsilon_N2(lam)
    c2  = np.cos(np.radians(theta))**2
    num = (180.+13.*eps) + (180.+eps)*c2
    den = (180.+52.*eps) + (180.+4.*eps)*c2
    return num/den

def f0_O2(lam, theta):
    '''
    lam in nm
    theta in deg
    '''
    eps = Epsilon_O2(lam)
    c2  = np.cos(np.radians(theta))**2
    num = (180.+13.*eps) + (180.+eps)*c2
    den = (180.+52.*eps) + (180.+4.*eps)*c2
    return num/den

# Joiner, J., Bhartia, P. K., Cebula, R. P., Hilsenrath, E., McPeters, R. D., & Park, H. (1995). Rotational Raman scattering (Ring effect) 
# in satellite backscatter ultraviolet measurements. Applied Optics, 34(21), 4513. doi:10.1364/ao.34.004513
## !!!! Erreur dans le papier original sur les coeffs de Placzek-Teller Anti Stokes !!!

def K(lam, theta):
    '''
    lam in nm
    theta in deg
    '''
    return (1.-f0_O2(lam,theta))/(1.-f0_N2(lam,theta))

def bjp(J):
    return 3.*(J+1)*(J+2)/2./(2*J+1)/(2*J+3)

def bjm(J):
    b = 3.*J*(J-1)/2./(2*J+1)/(2*J-1)
    b[J<=1] = 0.
    return b
    
def L_O2(T):
    '''
    O2 Rotational Raman Spectrum

    T in K
    '''
    B0 = 1.4378 # cm-1
    J  = np.linspace(0, 36, num=37, dtype=int)
    gj = np.array([1 if is_odd(j) else 0 for j in J])
    Ej = J*(J+1)*cst.h*cst.c*B0*100 # B0 translated in m-1 !!!
    Fj = gj*(2*J+1) * np.exp(-Ej/(cst.k*T))
    
    Lj_S  = Fj * bjp(J)
    Dnu_S = -(4*J+6)*B0
    not0  = Lj_S != 0.
    Lj_S  = Lj_S[not0]
    Dnu_S = Dnu_S[not0]
    Lj_A  = Fj * bjm(J)
    not0  = Lj_A != 0.
    Dnu_A =  (4*J-2)*B0
    Lj_A  = Lj_A[not0]
    Dnu_A = Dnu_A[not0]

    norm = Lj_S.sum() + Lj_A.sum()
    return Dnu_S, Lj_S/norm, Dnu_A, Lj_A/norm

def L_N2(T):
    '''
    N2 Rotational Raman Spectrum

    T in K
    '''
    B0 = 1.9897 # cm-1
    J  = np.linspace(0, 36, num=37, dtype=int)
    gj = np.array([3 if is_odd(j) else 6 for j in J])
    Ej = J*(J+1)*cst.h*cst.c*B0*100 # B0 translated in m-1 !!!
    Fj = gj*(2*J+1) * np.exp(-Ej/(cst.k*T))
    
    Lj_S  = Fj * bjp(J)
    Dnu_S = -(4*J+6)*B0
    not0  = Lj_S != 0.
    Lj_S  = Lj_S[not0]
    Dnu_S = Dnu_S[not0]
    Lj_A  = Fj * bjm(J)
    not0  = Lj_A != 0.
    Dnu_A =  (4*J-2)*B0
    Lj_A  = Lj_A[not0]
    Dnu_A = Dnu_A[not0]

    norm = Lj_S.sum() + Lj_A.sum()
    return Dnu_S, Lj_S/norm, Dnu_A, Lj_A/norm

def L(lam, theta, T):
    '''
    Air Rotational Raman Spectrum
    
    lam central wavelength in nm
    theta in deg
    T in K
    '''
    Dnu_S_N2, Lj_S_N2, Dnu_A_N2, Lj_A_N2 = L_N2(T)
    Dnu_S_O2, Lj_S_O2, Dnu_A_O2, Lj_A_O2 = L_O2(T)
    Lj_S_N2 *= X_N2
    Lj_A_N2 *= X_N2
    Lj_S_O2 *= X_O2*K(lam, theta)
    Lj_A_O2 *= X_O2*K(lam, theta)
    
    nu0 = 1e7/(lam) # nu0 in cm-1
    # compute output lamnda in nm
    lam_S_N2 = 1e7/(nu0+Dnu_S_N2)
    lam_A_N2 = 1e7/(nu0+Dnu_A_N2)
    lam_S_O2 = 1e7/(nu0+Dnu_S_O2)
    lam_A_O2 = 1e7/(nu0+Dnu_A_O2)
    
    norm = Lj_S_N2.sum() + Lj_A_N2.sum() + Lj_S_O2.sum() + Lj_A_O2.sum() 
    
    lam_out= np.concatenate([lam_A_N2, lam_A_O2, lam_S_N2, lam_S_O2])
    L_out  = np.concatenate([Lj_A_N2, Lj_A_O2, Lj_S_N2, Lj_S_O2])/norm
    ii = np.argsort(lam_out)
    
    # return spectrum with increasing wavelengths
    return lam_out[ii], L_out[ii]


def L2d(lam, theta, T):
    '''
    Air Rotational Raman Spectrum
    
    lam central wavelength in nm
    theta in deg
    T in K
    '''
    KK = K(lam, theta)
    nlam = lam.size
    Dnu_S_N2, Lj_S_N2, Dnu_A_N2, Lj_A_N2 = L_N2(T)
    Dnu_S_O2, Lj_S_O2, Dnu_A_O2, Lj_A_O2 = L_O2(T) 
    norm_N2 = Lj_S_N2.sum() + Lj_A_N2.sum()
    norm_O2 = Lj_S_O2.sum() + Lj_A_O2.sum()
    Lj_S_N2/=norm_N2
    Lj_A_N2/=norm_N2
    Lj_S_O2/=norm_O2
    Lj_A_O2/=norm_O2
    Lj_S_N2 = np.stack([Lj_S_N2]*nlam) * X_N2
    Lj_A_N2 = np.stack([Lj_A_N2]*nlam) * X_N2
    Lj_S_O2 = Lj_S_O2[np.newaxis, :] * X_O2 * KK[:, np.newaxis]
    Lj_A_O2 = Lj_A_O2[np.newaxis, :] * X_O2 * KK[:, np.newaxis]
    norm = np.sum(Lj_S_N2, axis=1) + np.sum(Lj_A_N2, axis=1) + np.sum(Lj_S_O2, axis=1) + np.sum(Lj_A_O2, axis=1)
    Lj_S_N2/=norm[:, np.newaxis]
    Lj_A_N2/=norm[:, np.newaxis]
    Lj_S_O2/=norm[:, np.newaxis]
    Lj_A_O2/=norm[:, np.newaxis]
    #we add also negative unity impulse at zero for removal of elastic
    #L_out  = np.concatenate([Lj_A_N2, Lj_A_O2, Lj_S_N2, Lj_S_O2, np.stack([np.array([0])]*nlam)], axis=1)
    #Dnu_out= np.concatenate([Dnu_A_N2, Dnu_A_O2, Dnu_S_N2, Dnu_S_O2, np.array([0.])])
    L_out  = np.concatenate([Lj_A_N2, Lj_A_O2, Lj_S_N2, Lj_S_O2], axis=1)
    Dnu_out= np.concatenate([Dnu_A_N2, Dnu_A_O2, Dnu_S_N2, Dnu_S_O2])
    
    
    # reorganization with lambda instead od Dnu and increasing order
    nu0 = 1e7/lam
    lam_out = 1e7/(nu0[:, np.newaxis]+Dnu_out[np.newaxis,:])
    ii  = np.argsort(lam_out, axis=1)
    lam_out = np.take_along_axis(lam_out, ii, axis=1)
    L_out   = np.take_along_axis(L_out,   ii, axis=1)
    
    return lam_out, L_out


def L2d_inv(lam, theta, T):
    '''
    Air Inverse Rotational Raman Spectrum
    
    lam central wavelength in nm
    theta in deg
    T in K
    '''
    
    Dnu_S_N2, Lj_S_N2, Dnu_A_N2, Lj_A_N2 = L_N2(T)
    Dnu_S_O2, Lj_S_O2, Dnu_A_O2, Lj_A_O2 = L_O2(T) 
    # reorganization with lambda instead od Dnu
    Dnu_in= np.concatenate([Dnu_A_N2, Dnu_A_O2, Dnu_S_N2, Dnu_S_O2])
    nu0 = 1e7/lam
    lam_in1 = 1e7/(nu0[:, np.newaxis]-Dnu_S_O2[np.newaxis,:])
    lam_in2 = 1e7/(nu0[:, np.newaxis]-Dnu_A_O2[np.newaxis,:])
    nlam = lam.shape[0]
    KK1 = K(lam_in1, theta)
    KK2 = K(lam_in2, theta)
    norm_N2 = Lj_S_N2.sum() + Lj_A_N2.sum()
    norm_O2 = Lj_S_O2.sum() + Lj_A_O2.sum()
    Lj_S_N2/=norm_N2
    Lj_A_N2/=norm_N2
    Lj_S_O2/=norm_O2
    Lj_A_O2/=norm_O2
    Lj_S_N2 = np.stack([Lj_S_N2]*nlam) * X_N2
    Lj_A_N2 = np.stack([Lj_A_N2]*nlam) * X_N2
    Lj_S_O2 = Lj_S_O2[np.newaxis, :] * X_O2 * KK1
    Lj_A_O2 = Lj_A_O2[np.newaxis, :] * X_O2 * KK2
    norm = np.sum(Lj_S_N2, axis=1) + np.sum(Lj_A_N2, axis=1) + np.sum(Lj_S_O2, axis=1) + np.sum(Lj_A_O2, axis=1)
    Lj_S_N2/=norm[:, np.newaxis]
    Lj_A_N2/=norm[:, np.newaxis]
    Lj_S_O2/=norm[:, np.newaxis]
    Lj_A_O2/=norm[:, np.newaxis]
    L_in    = np.concatenate([Lj_A_N2, Lj_A_O2, Lj_S_N2, Lj_S_O2], axis=1)

    lam_in = 1e7/(nu0[:, np.newaxis]-Dnu_in[np.newaxis,:])
    ii  = np.argsort(lam_in, axis=1)
    lam_in = np.take_along_axis(lam_in, ii, axis=1)
    L_in   = np.take_along_axis(L_in,   ii, axis=1)
    
    return lam_in, L_in



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# This module is deprecated. It will be removed in one of the next versions of smartg.
# All the classes/functions are now available in the geoclide package.

from smartg.geometry import Vector, Point, Ray
from smartg.geometry import Dot, Cross, Normalize, CoordinateSystem, \
    FaceForward, MaxDimension, Permute
from smartg.diffgeom import DifferentialGeometry
from smartg.transform import Transform
import math 
import matplotlib.pyplot as plt
import numpy as np

from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from warnings import warn


#####################################################################################
class Shape(object):
    '''
    Creation of the class Shape
    '''
    indShape = 0
    def __init__(self, ObjectToWorld, WorldToObject):
        warn_message = 'Using Sphere, Triangle or TriangleMesh from smartg.shape ' + \
                       'is depracated as of smartg v1.1.0. Use geoclide package instead.\n ' + \
                       'For example: import geoclide as gc; gc.Sphere(); gc.Triangle(); gc.TriangleMesh().'
        warn(warn_message, DeprecationWarning)
        Shape.indShape += 1
        self.oTw = ObjectToWorld
        self.wTo = WorldToObject
#####################################################################################

#####################################################################################
class Sphere(Shape):
    '''
    Creation of the class Sphere
    '''
    def __init__(self, oTw, wTo, rad, z0, z1, pm):
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.radius = rad
        self.zmin = Clamp(min(z0, z1), -self.radius, self.radius)
        self.zmax = Clamp(max(z0, z1), -self.radius, self.radius)
        self.thetaMin = math.acos(Clamp(self.zmin/self.radius, -1, 1))
        self.thetaMax = math.acos(Clamp(self.zmax/self.radius, -1, 1))
        self.phiMax = math.radians(Clamp(pm, 0, 360))
        Shape.indShape += 1

    def Intersect(self, r):

        ray = Ray(r)
        #print "MRAYY0", ray.o
        ray.o = self.wTo[r.o]
        ray.d = self.wTo[r.d]

        #print "MRAYY", ray.o
        self.thit = None; self.dg = None

        # Compute quadratic sphere coefficients
        A = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
        B = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
        C = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
            self.radius*self.radius

        # Solve quadratic equation for _t_ values
        exist = None; t0 = None; t1 = None;
        exist, t0, t1 = Quadratic(exist, t0, t1, A, B, C)
        if (exist == False): return False
        
        # Compute intersection distance along ray
        if (t0 > ray.maxt or t1 < ray.mint): return False
        self.thit = t0
        if (t0 < ray.mint):
            self.thit = t1
            if (self.thit > ray.maxt): return False

        # Compute sphere hit position and $\phi$
        phit = ray[self.thit]
        if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
        phi = math.atan2(phit.y, phit.x)
        if (phi < 0): phi += 2*math.pi

        #Test sphere intersection against clipping parameters
        if ((self.zmin > -self.radius and phit.z < self.zmin) or \
            (self.zmax <  self.radius and phit.z > self.zmax) or phi > self.phiMax):
            if (self.thit == t1): return False
            if (t1 > ray.maxt): return False
            self.thit = t1
            # Compute sphere hit position and $\phi$
            phit = ray[self.thit]
            if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
            phi = math.atan2(phit.y, phit.x)
            if (phi < 0): phi += 2*math.pi
            if ((self.zmin > -self.radius and phit.z < self.zmin) or \
                (self.zmax <  self.radius and phit.z > self.zmax) or \
                phi > self.phiMax):
                return False;

        #Find parametric representation of sphere hit
        u = phi / self.phiMax
        theta = math.acos(Clamp(phit.z / self.radius, -1, 1))
        v = (theta - self.thetaMin) / (self.thetaMax - self.thetaMin)

        # Compute sphere $\dpdu$ and $\dpdv$
        zradius = math.sqrt(phit.x*phit.x + phit.y*phit.y)
        invzradius = 1 / zradius
        cosphi = phit.x * invzradius
        sinphi = phit.y * invzradius
        dpdu = Vector(-self.phiMax * phit.y, self.phiMax * phit.x, 0);
        dpdv = Vector(Vector(phit.z * cosphi, phit.z * sinphi, -self.radius \
                        * math.sin(theta)) * (self.thetaMax-self.thetaMin))

        # Initialize _DifferentialGeometry_ from parametric information
        self.dg = DifferentialGeometry(self.oTw[phit], self.oTw[dpdu], self.oTw[dpdv], u, v, self)

        return True

    def Area(self):
        return (self.phiMax * self.radius * (self.zmax-self.zmin))
#####################################################################################

#####################################################################################
class Triangle(Shape):
    '''
    Creation of the class Sphere
    '''
    def __init__(self, oTw, wTo, a = Point(), b = Point(), c = Point()):
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.p1 = a; self.p2 = b; self.p3 = c;

    def Intersect(self, r):
        ray = Ray(r)
        e1 = self.p2 - self.p1
        e2 = self.p3 - self.p1
        s1 = Cross(ray.d, e2)
        divisor = Dot(s1, e1)

        if (divisor == 0):
            return False
        invDivisor = 1./divisor

        # compute the first barycentric coordinate
        s = ray.o - self.p1
        b1 = Dot(s, s1) * invDivisor
        if (b1 < -0.00000001 or  b1 > 1.00000001): #b1 < 0 or  b1 > 1):
            return False

        # compute the second barycentric coordinate
        s2 = Cross(s, e1)
        b2 = Dot(ray.d, s2) * invDivisor
        if (b2 < 0 or  b1+b2 > 1):
            return False

        # compute the time at the intersection point
        t = Dot(e2, s2) * invDivisor
        if (t < ray.mint or t > ray.maxt):
            return False

        # compute triangle partial derivatives
        uvs = np.array([[0., 0.], [1., 0.], [1., 1.]])

        # compute deltas for triangle partial derivatives
        du1 = uvs[0][0] - uvs[2][0]
        du2 = uvs[1][0] - uvs[2][0]
        dv1 = uvs[0][1] - uvs[2][1]
        dv2 = uvs[1][1] - uvs[2][1]
        dp1 = self.p1 - self.p3
        dp2 = self.p2 - self.p3
        determinant = du1 * dv2 - dv1 * du2

        if (determinant == 0):
            dpdu, dpdv = CoordinateSystem(Normalize(Cross(e2, e1)))
        else:
            invdet = 1./determinant
            # myV1 = dp1*dv2
            # myV2 = dp2*dv1
            # print(type(myV1), type(myV2))
            dpdu = ( dp1*dv2   - dp2*dv1) * invdet
            dpdv = (dp1*(-du2) + dp2*du1) * invdet
        
        # interpolate $(u,v)$ triangle parametric coordinates
        b0 = 1 - b1 - b2
        tu = b0*uvs[0][0] + b1*uvs[1][0] + b2*uvs[2][0]
        tv = b0*uvs[0][1] + b1*uvs[1][1] + b2*uvs[2][1]

        # fill the DifferentialGeometry and thit
        self.dg = DifferentialGeometry(ray[t], dpdu, dpdv, tu, tv, self)
        self.thit = t

        return True
#####################################################################################

#####################################################################################
class TriangleM(Shape):
    '''
    Creation of the class Sphere
    '''
    def __init__(self, oTw, wTo, a = Point(), b = Point(), c = Point()):
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.p1 = a; self.p2 = b; self.p3 = c;

    def Intersect_old(self, r):
        ray = Ray(r)

        p1t = self.p1 - ray.o
        p2t = self.p2 - ray.o
        p3t = self.p3 - ray.o

        kz = MaxDimension(  Vector( abs(ray.d.x), abs(ray.d.y), abs(ray.d.z) )  )
        
        kx = kz + 1
        if(kx == 3):
            kx = 0
        ky = kx + 1
        if(ky == 3):
            ky = 0

        d = Permute(ray.d, kx, ky, kz)
        p1t = Permute(p1t, kx, ky, kz)
        p2t = Permute(p2t, kx, ky, kz)
        p3t = Permute(p3t, kx, ky, kz)
        
        sx = (d.x*-1)/d.z
        sy = (d.y*-1)/d.z
        sz = 1./d.z
        
        p1t.x = p1t.x + (sx*p1t.z)
        p1t.y = p1t.y + (sx*p1t.z)
        p2t.x = p2t.x + (sx*p2t.z)
        p2t.y = p2t.y + (sx*p2t.z)
        p3t.x = p3t.x + (sx*p3t.z)
        p3t.y = p3t.y + (sx*p3t.z)
         
        e1 = (p2t.x * p3t.y) - (p2t.y * p3t.x)
        e2 = (p3t.x * p1t.y) - (p3t.y * p1t.x)
        e3 = (p1t.x * p2t.y) - (p1t.y * p2t.x)

        # if( (e1 < -0.01 or e2 < -0.01 or e3 < -0.01) and
        #     (e1 > 0.01 or e2 > 0.01 or e3 > 0.01) ):
        #     return False

        det = e1 + e2 + e3

        if (det == 0):
            return False

        p1t.z = p1t.z * sz
        p2t.z = p2t.z * sz
        p3t.z = p3t.z * sz

        tScaled = e1*p1t.z + e2*p2t.z + e3*p3t.z

        if ( det < 0 and (tScaled >= 0 or tScaled > ray.maxt*det) ):
            return False

        invDet = 1./det
        b1 = e1 * invDet
        b2 = e2 * invDet
        b3 = e3 * invDet

        t = tScaled * invDet

        # phit = (self.p1*b1) + (self.p2*b2)+ (self.p3*b3)

        # compute triangle partial derivatives
        uvs = np.array([[0., 0.], [1., 0.], [1., 1.]])

        # compute deltas for triangle partial derivatives
        du1 = uvs[0][0] - uvs[2][0]
        du2 = uvs[1][0] - uvs[2][0]
        dv1 = uvs[0][1] - uvs[2][1]
        dv2 = uvs[1][1] - uvs[2][1]
        dp1 = self.p1 - self.p3
        dp2 = self.p2 - self.p3
        determinant = du1 * dv2 - dv1 * du2

        if (determinant == 0):
            dpdu, dpdv = CoordinateSystem(Normalize(Cross(e2, e1)))
        else:
            invdet = 1./determinant
            # myV1 = dp1*dv2
            # myV2 = dp2*dv1
            # print(type(myV1), type(myV2))
            dpdu = ( dp1*dv2   - dp2*dv1) * invdet
            dpdv = (dp1*(-du2) + dp2*du1) * invdet

        # interpolate $(u,v)$ triangle parametric coordinates
        tu = b1*uvs[0][0] + b2*uvs[1][0] + b3*uvs[2][0]
        tv = b1*uvs[0][1] + b2*uvs[1][1] + b3*uvs[2][1]
            
        self.dg = DifferentialGeometry(ray[t], dpdu, dpdv, tu, tv, self)
        self.thit = t
        
        return True
        
    # TODO bellow must be corrected
    def Intersect(self, r):
        ray = Ray(r)
        e1 = self.p2 - self.p1
        e2 = self.p3 - self.p1
        s1 = Cross(ray.d, e2)
        divisor = Dot(s1, e1)

        if (divisor == 0):
            return False
        invDivisor = 1./divisor

        # compute the first barycentric coordinate
        s = ray.o - self.p1
        b1 = Dot(s, s1) * invDivisor
        # print("=====================")
        # print("b1=", b1)
        # print("s=", s)
        # print("s1=", s1)
        # print("Dot(s, s1)=", Dot(s, s1))
        # print("invDivisor", invDivisor)
        if (b1 < -0.00000001 or  b1 > 1.00000001):
            return False

        # compute the second barycentric coordinate
        s2 = Cross(s, e1)
        b2 = Dot(ray.d, s2) * invDivisor
        if (b2 < 0 or  b1+b2 > 1):
            return False

        # compute the time at the intersection point
        t = Dot(e2, s2) * invDivisor
        if (t < ray.mint or t > ray.maxt):
            return False

        # compute triangle partial derivatives
        uvs = np.array([[0., 0.], [1., 0.], [1., 1.]])

        # compute deltas for triangle partial derivatives
        du1 = uvs[0][0] - uvs[2][0]
        du2 = uvs[1][0] - uvs[2][0]
        dv1 = uvs[0][1] - uvs[2][1]
        dv2 = uvs[1][1] - uvs[2][1]
        dp1 = self.p1 - self.p3
        dp2 = self.p2 - self.p3
        determinant = du1 * dv2 - dv1 * du2

        if (determinant == 0):
            dpdu, dpdv = CoordinateSystem(Normalize(Cross(e2, e1)))
        else:
            invdet = 1./determinant
            myV1 = dp1*dv2
            myV2 = dp2*dv1
            # print(type(myV1), type(myV2))
            dpdu = ( dp1*dv2   - dp2*dv1) * invdet
            dpdv = (dp1*(-du2) + dp2*du1) * invdet
        
        # interpolate $(u,v)$ triangle parametric coordinates
        b0 = 1 - b1 - b2
        tu = b0*uvs[0][0] + b1*uvs[1][0] + b2*uvs[2][0]
        tv = b0*uvs[0][1] + b1*uvs[1][1] + b2*uvs[2][1]

        # fill the DifferentialGeometry and thit
        self.dg = DifferentialGeometry(ray[t], dpdu, dpdv, tu, tv, self)
        self.thit = t

        return True
#####################################################################################

#####################################################################################
class TriangleMesh(Shape):
    '''
    Creation of the class Sphere
    '''
    def __init__(self, oTw, wTo, vi, P):
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        if isinstance(vi, np.ndarray):
            self.VertexIndex = vi
            self.p = P
            self.ntris = int(len(np.atleast_1d(vi))/3)
            self.nverts = np.amax(self.VertexIndex) + 1
            self.reftri = [None]*self.ntris

            for i in range(0, self.nverts):
                self.p[i] = self.oTw[P[i]]
            
            for i in range(0, self.ntris):
                PointA = self.p[self.VertexIndex[3*i]]
                PointB = self.p[self.VertexIndex[3*i + 1]]
                PointC = self.p[self.VertexIndex[3*i + 2]]
                self.reftri[i] = TriangleM(Transform(), Transform(), PointA, PointB, PointC)
        else:
            raise NameError('error: vi or/and P are not numpy array***')

    def Intersect(self, r):
        self.dg = None
        self.thit = float("inf")
        for i in range(0, self.ntris):
            mybool = self.reftri[i].Intersect(r)
            if mybool:
                TriThit = self.reftri[i].thit
                if self.thit > TriThit:
                    self.thit = TriThit
                    self.dg = self.reftri[i].dg
        
        if self.dg == None:
            return False
            
        return True
#####################################################################################
def swap(a, b):
    warn_message = 'Using swap from smartg.shape is depracated as of smartg v1.1.0.'
    warn(warn_message, DeprecationWarning)
    return b, a

def Clamp(val, low, high):
    warn_message = 'Using Clamp from smartg.shape is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide or numpy package instead.\n ' + \
                   'For example: import geoclide as gc; gc.clamp();import numpy as np; np.clip()'
    warn(warn_message, DeprecationWarning)
    if (val < low): return low
    elif (val > high): return high
    else: return val

def Quadratic(exist, t0, t1, A, B, C):
    warn_message = 'Using Quadratic from smartg.shape is depracated as of smartg v1.1.0. ' + \
                   'Use geoclide package instead.\n ' + \
                   'For example: import geoclide as gc; gc.quadratic().'
    warn(warn_message, DeprecationWarning)
    #  Find quadratic discriminant
    discrim = (B * B) - (4 * A * C)
    if (discrim < 0): return False, None, None
    rootDiscrim = math.sqrt(discrim)

    # Compute quadratic _t_ values
    if (B < 0):
        q = -0.5 * (B - rootDiscrim)
    else:
        q = -0.5 * (B + rootDiscrim)
    if (A != 0):
        t0 = q / A
    else:
        t0 = C / q
    t1 = C / q
    if (t0 > t1): t0, t1 = swap(t0, t1)
    return True, t0, t1
#####################################################################################

#####################################################################################
if __name__ == '__main__':

    uvs = np.array([[0, 0], [1, 0], [1, 1]])
    print("uvs[0, 0] =", uvs[0, 0], "uvs[0, 1] =", uvs[0, 1])
    print("uvs[1, 0] =", uvs[1, 0], "uvs[1, 1] =", uvs[1, 1])
    print("uvs[2, 0] =", uvs[2, 0], "uvs[2, 1] =", uvs[2, 1])
    
    aVec = np.array(30)
    aVec2 = np.zeros(5, dtype=np.float32)
    aVec2[0] = 1
    aVec2[2] = 368
    aVec2[3] = 68
    print("lenght =", len(np.atleast_1d(aVec)), type(aVec))
    print("lenght =", len(np.atleast_1d(aVec2)), type(aVec2), "max=", np.amax(aVec2), "min=",np.amin(aVec2))
 
    ARAY = Ray(Point(0, 0, 10), Vector(-0.3, 0.5, -1), 0, 0, 30)
    ARAY2 = Ray(Point(0, 0, 10), Vector(-1.8, 1.9, -1), 0, 0, 30)
    vi = np.array([0, 1, 2,
                   2, 3, 1], dtype=np.int32)
    P = np.array([0,   0,  0,
                  -30, 0,  0,
                  0,   30, 0,
                  -30, 30, 0], dtype=np.float32)
    trimesh = TriangleMesh(Transform(), Transform(), vi, P)
    bool2 = trimesh.Intersect(ARAY2)
    if bool2:
        print("There is an intersection with the triangle Mesh***")
        print("At p = (", trimesh.dg.p.x, ", ", trimesh.dg.p.y, ", ", trimesh.dg.p.z, ")")
        tHitmesh = trimesh.thit
    else:
        print("No intersection with the triangle Mesh***")


    A1 = Point(0, 0, 0) ; A2 = Point(-30, 0, 0); A3 = Point(0, 30, 0);
    tri =  Triangle(Transform(), Transform(), A1, A2, A3)
    bool1 = tri.Intersect(ARAY)
    if bool1:
        print("There is an intersection with the triangle***")
        print("At p = (", tri.dg.p.x, ", ", tri.dg.p.y, ", ", tri.dg.p.z, ")")
        tHit = tri.thit
    else:
        print("No intersection with the triangle***")

    # transform needed for the first sphere
    TSph1 = Transform()

    TRX = TSph1.rotateX(90)
    TSph1 = TRX
    invTSph1 = TSph1.inverse(TSph1)

    # create the first sphere
    Sph1 = Sphere(TSph1, invTSph1, 60, -60, 60, 29.9)
    # ================================================
    # transform needed for the second sphere
    TSph2 = Transform()

    TTrans = TSph2.translate(Vector(-110, 0, 0))
    TSph2 = TTrans
    invTSph2 = TSph2.inverse(TSph2)

    # create the second sphere
    Sph2 = Sphere(TSph2, invTSph2, 50, -50, 50, 360)


    # create the matplotlib figure
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    # parameters of the first sphere
    u1 = np.linspace(0, Sph1.phiMax, 20)
    v1 = np.linspace(Sph1.thetaMin, Sph1.thetaMax, 20)
    myPoint1 = TSph1[Point(Sph1.radius * np.outer(np.cos(u1), np.sin(v1)), Sph1.radius * np.outer(np.sin(u1), np.sin(v1)), Sph1.radius * np.outer(np.ones(np.size(u1)), np.cos(v1)))]
    x1 = myPoint1.x
    y1 = myPoint1.y
    z1 = myPoint1.z

    # parameters of the second sphere
    u2 = np.linspace(0, Sph2.phiMax, 20)
    v2 = np.linspace(Sph2.thetaMin, Sph2.thetaMax, 20)
    myPoint2 = TSph2[Point(Sph2.radius * np.outer(np.cos(u2), np.sin(v2)), Sph2.radius * np.outer(np.sin(u2), np.sin(v2)), Sph2.radius * np.outer(np.ones(np.size(u2)), np.cos(v2)))]
    x2 = myPoint2.x
    y2 = myPoint2.y
    z2 = myPoint2.z

    # Photon paramaters plot thanks to straight equation (SunZ = A*SunX + b)
    # where SunY is constant and equal to 0
    tan_S = np.tan(60*(np.pi / 180.))      # zenith angle at 60°
    SunX = np.linspace(0, 120*tan_S, 100)
    xs = SunX
    ys = 0 * SunX                          # multiply by SunX to get an array
    zs = (1./tan_S) * SunX

    # Triangles parameters
    # v1 = [0, 0, 0]
    # v2 = [-30, 0, 0]
    # v3 = [0, 30, 0]

    # x = [v1[0], v2[0], v3[0]]
    # y = [v1[1], v2[1], v3[1]]
    # z = [v1[2], v2[2], v3[2]]
    # verts = [zip(x, y, z)]
    # ax.add_collection3d(Poly3DCollection(verts, facecolors='orange', linewidths=2))

    # Triangles mesh parameters
    TrV1 = [None]*trimesh.ntris
    TrV2 = [None]*trimesh.ntris
    TrV3 = [None]*trimesh.ntris
    Trx = [None]*trimesh.ntris
    Try = [None]*trimesh.ntris
    Trz = [None]*trimesh.ntris
    for i in range(0, trimesh.ntris):
        print("p1 = (", trimesh.reftri[i].p1.x, ", ", trimesh.reftri[i].p1.y, ", ", trimesh.reftri[i].p1.z, ")")
        print("p2 = (", trimesh.reftri[i].p2.x, ", ", trimesh.reftri[i].p2.y, ", ", trimesh.reftri[i].p2.z, ")")
        print("p3 = (", trimesh.reftri[i].p3.x, ", ", trimesh.reftri[i].p3.y, ", ", trimesh.reftri[i].p3.z, ")")
        TrV1[i] = [trimesh.reftri[i].p1.x, trimesh.reftri[i].p1.y, trimesh.reftri[i].p1.z]
        TrV2[i] = [trimesh.reftri[i].p2.x, trimesh.reftri[i].p2.y, trimesh.reftri[i].p2.z]
        TrV3[i] = [trimesh.reftri[i].p3.x, trimesh.reftri[i].p3.y, trimesh.reftri[i].p3.z]
        Trx[i] = [TrV1[i][0], TrV2[i][0], TrV3[i][0]]
        Try[i] = [TrV1[i][1], TrV2[i][1], TrV3[i][1]]
        Trz[i] = [TrV1[i][2], TrV2[i][2], TrV3[i][2]]
        Trverts = [list(zip(Trx[i], Try[i], Trz[i]))]
        Poly = Poly3DCollection(Trverts, facecolors='orange', linewidths=2)
        Poly.set_alpha(0.1)
        ax.add_collection3d(Poly)



    if bool1:
        t3 = np.linspace(ARAY.mint, tHit, 100)
        x3 = ARAY.o.x + t3*ARAY.d.x
        y3 = ARAY.o.y + t3*ARAY.d.y
        z3 = ARAY.o.z + t3*ARAY.d.z
        ax.plot(x3, y3, z3, color='r', linewidth=2)
        tnn = np.linspace(0, 1, 20)
        P1 = tri.dg.p ; N1 = tri.dg.nn;
        N1 = FaceForward(N1, ARAY.d * -1)
        x5 = P1.x + tnn * N1.x
        y5 = P1.y + tnn * N1.y
        z5 = P1.z + tnn * N1.z
        ax.plot(x5, y5, z5, color='g', linewidth=4)

    if bool2:
        tm1 = np.linspace(ARAY2.mint, tHitmesh, 100)
        xm1 = ARAY2.o.x + t3*ARAY2.d.x
        ym1 = ARAY2.o.y + t3*ARAY2.d.y
        zm1 = ARAY2.o.z + t3*ARAY2.d.z
        ax.plot(xm1, ym1, zm1, color='r', linewidth=2)
        tnnmesh = np.linspace(0, 1, 20)
        Pmesh = trimesh.dg.p ; Nmesh = trimesh.dg.nn;
        Nmesh = FaceForward(Nmesh, ARAY.d * -1)
        xm2 = Pmesh.x + tnnmesh * Nmesh.x
        ym2 = Pmesh.y + tnnmesh * Nmesh.y
        zm2 = Pmesh.z + tnnmesh * Nmesh.z
        ax.plot(xm2, ym2, zm2, color='g', linewidth=4)

    # plot all the geometries
    ax.scatter([0],[0],[0],color="g",s=140)
    ax.plot_surface(x1, y1, z1, rstride=1, cstride=1, color='b')
    ax.plot_surface(x2, y2, z2, rstride=1, cstride=1, color='b')
    ax.plot(xs, ys, zs, color='y', linewidth=4)
    ax.set_xlim3d(-120, 120)
    ax.set_ylim3d(-120, 120)
    ax.set_zlim3d(-120, 120)
    # Show the geometries
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from smartg.atmosphere import AtmAFGL, AerOPAC
from pathlib import Path
import logging

# ***************************** Global variable(s) ******************************
ROOTPATH = Path(__file__).resolve().parent.parent.parent

AER_SPHERIC = ['continental_clean','continental_average','continental_polluted','urban','desert_spheric',
               'maritime_clean','maritime_polluted','maritime_tropical','arctic','antarctic_spheric']

# OPAC 1998 paper table 3
AER_SPHERIC_OPAC = {}
AER_SPHERIC_OPAC.update({'continental_clean':{'tau':0.064, 'ssa':0.972}})
AER_SPHERIC_OPAC.update({'continental_average':{'tau':0.151, 'ssa':0.925}})
AER_SPHERIC_OPAC.update({'continental_polluted':{'tau':0.327, 'ssa':0.892}})
AER_SPHERIC_OPAC.update({'urban':{'tau':0.643, 'ssa':0.817}})
AER_SPHERIC_OPAC.update({'desert_spheric':{'tau':0.286, 'ssa':0.888}})
AER_SPHERIC_OPAC.update({'maritime_clean':{'tau':0.096, 'ssa':0.997}})
AER_SPHERIC_OPAC.update({'maritime_polluted':{'tau':0.117, 'ssa':0.975}})
AER_SPHERIC_OPAC.update({'maritime_tropical':{'tau':0.056, 'ssa':0.998}})
AER_SPHERIC_OPAC.update({'arctic':{'tau':0.063, 'ssa':0.887}})
AER_SPHERIC_OPAC.update({'antarctic_spheric':{'tau':0.072, 'ssa':1.000}})

DELTA_R_TAU = 0.6 # relative error in %
DELTA_R_SSA = 0.2 # relative error in %
# *******************************************************************************

# *********************************** logging ***********************************
# Create log file
Path(ROOTPATH / 'smartg' / 'tests' / 'logs').mkdir(parents=True, exist_ok=True)

# Create a named logger
logger = logging.getLogger('test_OPAC')
logger.setLevel(logging.INFO)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.ERROR)

# Set the formatter for the console handler
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s',
datefmt='%m/%d/%Y %I:%M:%S%p')
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)

# Create a file handler
file_handler = logging.FileHandler(ROOTPATH / 'smartg' / 'tests' / 'logs' / 'OPAC.log', mode='w')
file_handler.setLevel(logging.INFO)

# Set the formatter for the file handler
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%m/%d/%Y %I:%M:%S%p')
file_handler.setFormatter(formatter)

# Add the file handler to the logger
logger.addHandler(file_handler)
# *******************************************************************************


@pytest.mark.parametrize('mod', AER_SPHERIC)
def test_aer_spheric(mod):
    wl_ref = 550.
    aer_comp = AerOPAC(mod, tau_ref=None, w_ref=wl_ref, rh_mix=80., rh_free=0., rh_stra=0.)
    pro = AtmAFGL('afglt', comp=[aer_comp]).calc(wl_ref)

    opac_tau = AER_SPHERIC_OPAC[mod]['tau']
    opac_ssa = AER_SPHERIC_OPAC[mod]['ssa']

    cond_tau1 = pro['OD_p'][0,-1] > (opac_tau - (opac_tau*DELTA_R_TAU*1e-2 + pro['OD_p'][0,-1]*DELTA_R_TAU*1e-2))
    cond_tau2 = pro['OD_p'][0,-1] < (opac_tau + (opac_tau*DELTA_R_TAU*1e-2 + pro['OD_p'][0,-1]*DELTA_R_TAU*1e-2))
    cond_ssa1 = pro['ssa_p_atm'][0,-1] > (opac_ssa - (opac_ssa*DELTA_R_SSA*1e-2 + pro['ssa_p_atm'][0,-1]*DELTA_R_SSA*1e-2))
    cond_ssa2 = pro['ssa_p_atm'][0,-1] < (opac_ssa + (opac_ssa*DELTA_R_SSA*1e-2 + pro['ssa_p_atm'][0,-1]*DELTA_R_SSA*1e-2))

    logger.info(f"{mod} - tau_ref={AER_SPHERIC_OPAC[mod]['tau'] :.3f} - tau_calc={pro['OD_p'][0,-1] :.3f}")
    logger.info(f"{mod} - ssa_ref={AER_SPHERIC_OPAC[mod]['ssa'] :.3f} - ssa_calc={pro['ssa_p_atm'][0,-1] :.3f}")

    assert ( cond_tau1 and cond_tau2 ), f"Problem with {mod} tau value, get {pro['OD_p'][0,-1]:.5f} " + \
        f"instead of {opac_tau:.5f} +- {(opac_tau*DELTA_R_TAU*1e-2 + pro['OD_p'][0,-1]*DELTA_R_TAU*1e-2):.5f}"
    
    assert ( cond_ssa1 and cond_ssa2 ), f"Problem with {mod} tau value, get {pro['ssa_p_atm'][0,-1]:.5f} " + \
        f"instead of {opac_ssa:.5f} +- {(opac_ssa*DELTA_R_SSA*1e-2 + pro['ssa_p_atm'][0,-1]*DELTA_R_SSA*1e-2):.5f}"



import numpy as np
import jax.numpy as jnp
from jax import value_and_grad, vmap, jit
import xarray
import jax

def get_histories(m, LEVEL=0, IDIR=0,verbose=False):
    ''' 
    Return photons histories main outputs
    
    Input
        m : a MLUT (or xarray) SMART-G output with the ALIS option and hist=True having been set
        
    Keyword 
        LEVEL : 0 or 1 (up TOA or down 0+ levels only)
        verbose : print the Number of injected photons (N), 
                  Number of Local Estimate virtual photons (NLE), 
                  Number of Low Resolution wavelengths recorded (NLR)
                  Number of vertical layers (NL)
        
    Output
        a tuple consisting of 
            N : the number of injected photons
            S : A ndarray of size (NLE, 4) for 4 Stokes components
            D : A ndarray of size (NLE, NL) for cumulative distances traveled in layers
            w : A ndarray of size (NLE, NLR) for corrective scattering weights for the different LR wavelengths
            nrrs : A ndarray of size (NLE) of Rotational Raman Scattering event flag (1 : RRS, 0: no RRS)
            nref : A ndarray of size (NLE) of number of reflection on the surface (as described by the keyword surf in the run method)
            nrrs : A ndarray of size (NLE) of Sun Induced Fluorescence event flag (1 : SIF, 0: no SIF)
            nvrs : A ndarray of size (NLE) of Vibrational Raman Scattering event flag (1 : VRS, 0: no VRS)
            nenv : A ndarray of size (NLE) of reflection on the environement (as described by the keyword env in the run method)
            //ith  : A ndarray of size (NLE) of index of the Zenith angle LE direction of the virtual photon
            nint : A ndarray of size (NLE) of number of reflection or scattering
    '''
    NL=m.axis('z_atm').size-1 if not isinstance(m, xarray.Dataset) else m['z_atm'].size-1
    tabHist_ = np.squeeze(m['histories'].data)
    tabHist = tabHist_[LEVEL, :,:]
    if verbose : print (tabHist.shape)
    w0      = tabHist[:, NL+4:-6] 
    #D0      = tabHist[:,0]
    good    = w0[:,0]!=0
    ngood   = np.sum(good)
    N = m['Nphotons_in'].data[0,0]
    ###################
    S       = np.zeros((ngood,4),dtype=np.float32) 
    D       = tabHist[good,     :NL  ]
    S[:,:4] = tabHist[good, NL:NL+4  ]
    w       = tabHist[good, NL+4:-6  ]
    nrrs    = tabHist[good,      -6  ]
    nref    = tabHist[good,      -5  ]
    nsif    = tabHist[good,      -4  ]
    nvrs    = tabHist[good,      -3  ]
    nenv    = tabHist[good,      -2  ]
    #ith     = tabHist[good,      -1  ]
    nint     = tabHist[good,      -1  ]
    #
    if verbose : print('Number of photons in : {}\nNumber of LE photons : {}\nNumber of LR wavelengths : {}\nNumber of Layers : {}'.format(N, *w.shape, NL))
    
    return N, S, D, w, nrrs, nref, nsif, nvrs, nenv, nint




def Si(lam, kabs, alb, sik, wi_lr, Dij, Ki, lam_lr_grid):
    '''
    JAX based computation ONE Stoke component of ONE virtual photon for ONE High Resolution wavelength
    
    Input
        lam : current HR wavelength (nm)
        kabs: A ndarray of size (NL) of gaseous absorption coefficient for the current wavelength and for all layers
        alb: surface albedo for the current wavelength
        sik : virtual LE photons Stokes component k
        wi_lr: A ndarray of size (NLR) virtual LE photons corrective scattering weights for the different LR wavelengths
        Dij  : A ndarray of size (NL) of the virtual LE photons cumulative distances traveled in layers
        Ki   : Number of reflection on the surface
        lam_lr_grid : A ndarray of size (NLR) LR wavelengths grid
    '''
    # interpolation of scattering weights at low spectral resolution to current lambda
    wi = jnp.interp(lam, lam_lr_grid, wi_lr)
    
    return sik * wi * jnp.exp(- jnp.sum(Dij * kabs)) * alb**Ki
    
    
def Si2(lam, kabs, alb, sik, wi_lr, Dij, Ki, lam_lr_grid):
    '''
    JAX based computation of the square of ONE Stoke component of ONE virtual photon for ONE High Resolution wavelength
    
    Input
        lam : current HR wavelength (nm)
        kabs: A ndarray of size (NL) of gaseous absorption coefficient for the current wavelength and for all layers
        alb: surface albedo for the current wavelength
        sik : virtual LE photons Stokes component k
        wi_lr: A ndarray of size (NLR) virtual LE photons corrective scattering weights for the different LR wavelengths
        Dij  : A ndarray of size (NL) of the virtual LE photons cumulative distances traveled in layers
        Ki   : Number of reflection on the surface
        lam_lr_grid : A ndarray of size (NLR) LR wavelengths grid
    '''
    return Si(lam, kabs, alb, sik, wi_lr, Dij, Ki, lam_lr_grid)**2



def BigSum(S, grad=None, only_I=False):
    '''
    JAX based function for computing ALL the Stokes vectors for ALL High Resolution wavelengths and for ALL LE photons
    '''
    if grad is not None : S = value_and_grad(S, argnums=grad)
    f1m = vmap(S,  in_axes=(0   ,    0,    0, None, None, None, None, None))  # co varying wavelengths inputs
    f2m = vmap(f1m,in_axes=(None, None, None,    0,    0,    0,    0, None)) # co vaying LE photons inputs
    f3m = vmap(f2m,in_axes=(None, None, None,    1, None, None, None, None)) # co varying Stoke components inputs

    if only_I : return jit(f2m)
    else : return jit(f3m)


#!/usr/bin/env python
# vim:fileencoding=utf-8


'''
Calculate the ocean water absorption, scattering coefficients and phase
function using the wavelength and the chlorophyll concentration
'''

from phase_functions import PhaseFunction
from iop import IOP
from numpy import cos, pi, array
from numpy import arange, zeros
import numpy as np


wl_1 = array([350.0525, 450.0666, 550.084, 650.099])

aw   = array([0.0204, 0.0092, 0.0565, 0.3400])
bw   = array([0.0134, 0.0045, 0.0019, 0.0010])

ablk = array([0.0204, 0.0092, 0.3400, 0.6240])
wblk = array([0.3964, 0.3285, 0.0325, 0.0029])


class IOP_AOS_WATER(IOP):
    '''

    Parameters:
        NANG: number if angles for the phase function
        ang_trunc: truncation angle in degrees
        ALB: albedo of the sea floor
        DEPTH: depth in m ; default None = Semi infinite
        pfwav: list of arrays at which the phase functions are calculated
    '''
    def __init__(self, NANG=72001, ALB=0., DEPTH=None, pfwav=None, verbose=False):
        self.NANG = NANG
        self.alb = ALB
        self.depth = DEPTH
        self.pfwav=pfwav
        self.verbose = verbose
        self.last = None   # last parameters (wl, profil, phases) , to avoid
                           # reprocessing wavelengths that have been calxculated already

    def calc(self, wl, skip_phase=False):
        '''
        Calculate atot, btot and phase function (monochromatic)

        Arguments:
            wl: wavelength in nm

        Returns (atot, [(b0, P0), (b1, P1), ...]) where:
            * atot is the total absorption coefficient
            * bi is the scattering coefficient (without truncation) of the ith component
            * Pi is the phase function of the ith component
        '''
        NANG = self.NANG

        #
        # wavelength index
        #
        if (wl < wl_1[0]) or (wl > wl_1[-1]):
            raise Exception('Error, wavelength {} is out of range ({}, {})'.format(wl, wl_1[0], wl_1[-1]))
        i1 = int((wl - wl_1[0])/(wl_1[1] - wl_1[0]))

        #
        # pure water coefficients
        #
        #a0 = ablk[i1]
        #b0 = a0*wblk[i1]/(1.-wblk[i1])
        a0 = aw[i1]
        b0 = bw[i1]
        #
        # phase function
        #
        if skip_phase:
            P0 = None
            P1 = None
        else:
            ang = pi * arange(NANG, dtype='float64')/(NANG-1)    # angle in radians
            # pure water
            pf0 = zeros((NANG, 4), dtype='float64') # pure water phase function with 0 depol factor
            pf1 = zeros((NANG, 4), dtype='float64')
            pf0[:,1] = 0.75
            pf0[:,0] = 0.75 * cos(ang)**2
            pf0[:,2] = 0.75 * cos(ang)
            pf0[:,3] = 0.
            P0 = PhaseFunction(ang, 2*pf0, degrees=False)
            #P0 = PhaseFunction(ang, pf0, degrees=False)
            P1 = PhaseFunction(ang, 2*pf1, degrees=False)
            #P1 = PhaseFunction(ang, pf1, degrees=False)


        return a0, [(b0, P0), (0., P1)]

    def __str__(self):
        return 'CHL={}'.format(self.chl)



if __name__ == '__main__':

    mod = IOP_AOS_WATER(pfwav=[350., 450., 550., 650.])
    # print mod.calc(500.)
    print mod.calc_bands(np.linspace(350, 650, 4))
